using System;
using System.Collections;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Microsoft.Samples.DirectX.UtilityToolkit;

namespace Demo
{
    /// <summary>
    /// This class will maintain the bullets used by the game
    /// </summary>
    public class AIBullet : IMoveableObject
    {
        #region Constants
        public const float BulletMass = 8.0f; // The size of the bullets
        private const int SlicesStacks = 4; // How 'complex' the bullets mesh is
		private const float GravityForce = 1f;
        //private const float GravityForce = -9.8f * BulletMass * 5; // The force of gravity
        private const float LifeTime = 3.0f; // Amount of time the bullet will 'stay alive'
        public const float BulletStrenth = 1800.0f; // Strength of the bullets being fired
		public const int damage = 20;
        #endregion

        #region Static Data (Mesh)
        private static Mesh bulletMesh = null;
        #endregion

        #region Instance Data
        private Vector3 pos; // Current position of the bullet
        private Vector3 vel; // Velocity of the bullet
        private float aliveTime = 0.0f; // Amount of time the bullet has been alive
        private AI owner = null; // The tank that fired this bullet
        private Material alphaMaterial; // Current material of the bullet
        #endregion

        #region Creation
        /// <summary>
        /// Create a new instance of the sky box class
        /// </summary>
        public AIBullet(Device device, Vector3 initialPos, Vector3 initialVel, AI parent)
        {
            // First create the mesh if it doesn't exist yet
            if (bulletMesh == null)
            {
                bulletMesh = Mesh.Sphere(device, BulletMass, SlicesStacks, SlicesStacks);
            }

            // Create the material for the bullets
            alphaMaterial = new Material();
            alphaMaterial.AmbientColor = new ColorValue(1.0f, 1.0f, 1.0f, 1.0f);
            alphaMaterial.DiffuseColor = new ColorValue(1.0f, 1.0f, 1.0f, 1.0f - (aliveTime /LifeTime) );

            // Store the position and velocity of the bullet
            pos = initialPos;
            vel = initialVel;

            // Store the parent
            owner = parent;
        }

        #endregion

        #region Cleanup
        /// <summary>
        /// Will be used to cleanup objects
        /// </summary>
        public static void Cleanup()
        {
            if (bulletMesh != null)
            {
                bulletMesh.Dispose();
                bulletMesh = null;
            }
        }
        #endregion

        #region Hit Testing
        /// <summary>
        /// Does a hit test on the object
        /// </summary>
        public bool HitTest(IMoveableObject obj)
        {
            // Don't do hit testing if you fired the bullet
            if (obj == owner)
            {
                return false;
            }
            // The hit testing for this will be simpler, simply compare the distance
            Vector3 diff = this.Position - obj.Position;
            return diff.Length() < (this.Radius + obj.Radius);
        }
        #endregion

        #region Properties
        /// <summary>
        /// Will be used to determine if the bullets are still alive
        /// </summary>
        public bool IsAlive
        {
            get { return aliveTime <= LifeTime; }
        }

		public int Damage
		{
			get { return damage;}
		}
        #endregion

        #region IMoveableObject Members

        /// <summary>
        /// The bullet's radius
        /// </summary>
        public float Radius
        {
            get { return BulletMass; }
        }

        /// <summary>
        /// The bullet's position
        /// </summary>
        public Vector3 Position
        {
            get { return pos; }
        }

        /// <summary>
        /// Update the bullet based on the time since the last frame
        /// </summary>
        public void Update(float elapsedTime)
        {
            // Update the bullet's lifetime
            aliveTime += elapsedTime;
            // And alpha mode
            alphaMaterial.DiffuseColor = new ColorValue(1.0f, 1.0f, 1.0f, 1.0f - (aliveTime /LifeTime) );

            // Store the initial position of the bullet
            Vector3 initialPosition = pos;

            // Velocity needs to scale by gravity first, gravity only affects the y axis
            vel.Y += GravityForce * elapsedTime;

            // Update the bullet's position
            pos += Vector3.Scale(vel, elapsedTime);

            // We may need to reflect our direction if there's a collision, store the points
//            Plane reflectionPlane;
//            if (currentLevel.HitTest(this, out reflectionPlane))
//            {
//                Matrix reflectionMatrix = Matrix.Zero;
//                reflectionMatrix.Reflect(reflectionPlane);
//
//                // Get the current velocity
//                float velocity = vel.Length();
//                // Reflect the velocity and the position
//                Vector3 reflectedVelocity = Vector3.TransformCoordinate(Direction, reflectionMatrix);
//                Vector3 reflectedPosition = Vector3.TransformCoordinate(initialPosition, reflectionMatrix);
//
//                // Update the velocity with the refected one
//                vel = Vector3.Scale(reflectedVelocity, velocity);
//                // Move the position to the reflected one, and add velocity back in
//                pos = reflectedPosition + Vector3.Scale(vel, elapsedTime);
//            }
        }

        /// <summary>
        /// The direction the bullets is currently travelling
        /// </summary>
        public Vector3 Direction
        {
            get { return Vector3.Normalize(vel); }
        }

        /// <summary>
        /// Sets the device up for rendering the bullets
        /// </summary>
        public static void SetBulletRenderStates(GameEngine engine)
        {
            // Turn on alpha blending
            engine.RenderingDevice.RenderState.AlphaBlendEnable = true;
            engine.RenderingDevice.RenderState.SourceBlend = Blend.SourceAlpha;
            engine.RenderingDevice.RenderState.DestinationBlend = Blend.InvSourceAlpha;
        }
        /// <summary>
        /// Resets the device back to 'defaults'
        /// </summary>
        public static void EndBulletRenderStates(GameEngine engine)
        {
            // Turn off alpha blending
            engine.RenderingDevice.RenderState.AlphaBlendEnable = false;
        }

        /// <summary>
        /// Render the bullet
        /// </summary>
        public void Draw(GameEngine engine, Camera c)
        {
            // Set the world transform
            engine.SetWorldTransform(Matrix.Translation(pos));
            
            // Render the bullet mesh if it's viewable
            if (c.ObjectInFrustum(this))
            {
        #if (DEBUG)
                engine.numberFaces += bulletMesh.NumberFaces;
                engine.numberVerts += bulletMesh.NumberVertices;
        #endif
                engine.RenderingDevice.SetTexture(0, null);
                engine.RenderingDevice.Material = alphaMaterial;
                bulletMesh.DrawSubset(0);
            }
        }

        #endregion
    }

    /// <summary>
    /// A collection of bullets
    /// </summary>
    public class AIBullets : IEnumerable
    {
        #region Instance Data
        private ArrayList bulletList = new ArrayList();
        #endregion

        #region Collection Methods
        /// <summary>
        /// Indexer for bullets
        /// </summary>
        public AIBullet this[int index]
        {
            get { return bulletList[index] as AIBullet; }
            set { bulletList[index] = value; }
        }

        /// <summary>
        /// Add a new bullet to the list
        /// </summary>
        public void Add(AIBullet b)
        {
            bulletList.Add(b);
        }
        /// <summary>
        /// Remove a bullet to the list
        /// </summary>
        public void Remove(AIBullet b)
        {
            bulletList.Remove(b);
        }

        /// <summary>
        /// Remove all bullets from the list
        /// </summary>
        public void Clear()
        {
            bulletList.Clear();
        }

        /// <summary>
        /// The number of items in the list
        /// </summary>
        public int Count
        {
            get { return bulletList.Count; }
        }
        #endregion

        #region IEnumerable Members
        /// <summary>
        /// Returns the enumerator for the bullet list
        /// </summary>
        public IEnumerator GetEnumerator()
        {
            return bulletList.GetEnumerator();
        }
        #endregion
    }
}