using System;
using System.Collections;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Input=Microsoft.DirectX.DirectInput;

namespace Demo
{
	/// <summary>
	/// This class will maintain a player in the game
	/// </summary>
	/// 



	public class AI : IDisposable, IMoveableObject
	{
		#region Constants
		private const float MaximumMovementSpeed = 600.0f;
		private const float MaximumRotationSpeed = 1.15f;
		private const float MaximumClimbSpeed = 1.15f;
	
		private const float GunCoolDown = 0.9f;
		private const float scanRange = 60000.0f;
		private const float approachRange = 1000f;
		private const float fireRange = 5000.0f;
		private const int   bullError = 500;
		#endregion

		#region Instance Data
		private Ship myShip = null;
		private IMoveableObject target = null;

		// Ship firing items
		public event EventHandler Fired;
		private float lastFiredTime = -GunCoolDown; // So you can fire right away
		private float totalTime = 0.0f;

		// Store the players name and allow it to be rendered
		private string playerName = string.Empty;
		private Font playerFont = null;
		private Sprite playerFontSprite = null;
		private int armor = 100;
		private Random random = new Random();
		
		public enum AI_STATE
		{
			PATROL,
			ATTACK

		}

		public int Armor
		{
			get { return armor;}
			set { armor = value;}
		}

		private AI_STATE ai_state;

		public AI_STATE CurrentState
		{
			get { return ai_state;}
			set { ai_state = value;}
		}
	
		/// <summary>
		/// The Ships current position
		/// </summary>
		public Vector3 Position
		{
			get { return myShip.Position; }
			set { myShip.Position = value; }
		}
		/// <summary>
		/// The direction the player is currently facing
		/// </summary>
		public Vector3 Direction
		{
			get { return myShip.ShipDirection; }
		}

		public float Orientation
		{
			get { return myShip.ShipOrientation; }
			set { myShip.ShipOrientation = value; }
		}

		/// <summary>
		/// The current rotation of the Ship
		/// </summary>
		public float Rotation
		{
			get { return myShip.ShipRotation; }
			set { myShip.ShipRotation = value; }
		}
		/// <summary>
		/// Radius of the player model
		/// </summary>
		public float Radius
		{
			get { return myShip.Radius; }
		}

		/// <summary>
		/// Name of the ship
		/// </summary>
		public string Name
		{
			get { return playerName; }
			set { playerName = value; }
		}

	    
		#endregion

		#region Creation
		/// <summary>
		/// Create a new player instance
		/// </summary>
		/// <param name="device">Device used for rendering</param>
		/// <param name="parent">Parent control used to hook events</param>
		public AI(Device device, GameEngine parent, IntPtr parentHandle, string name, bool isLocal)
		{
			// Create a new Ship
			myShip = new Ship(device);

			// Store the player's name
			playerName = name;
			float fontSize = 36.0f;

			// Create the font for the player
			playerFont = new Font(device, new System.Drawing.Font(
				"Arial", fontSize, System.Drawing.FontStyle.Bold | 
				System.Drawing.FontStyle.Italic));
			playerFontSprite = new Sprite(device);

			// Hook the font's events
			device.DeviceLost += new EventHandler(OnDeviceLost);
			device.DeviceReset += new EventHandler(OnDeviceReset);
			
			ai_state = AI_STATE.PATROL;

		}

		public AI(Device device, GameEngine parent, IntPtr parentHandle, string name, bool isLocal, string MeshFile)
		{
			// Create a new Ship
			myShip = new Ship(device, MeshFile);

			// Store the player's name
			playerName = name;
			float fontSize = 36.0f;

			// Create the font for the player
			playerFont = new Font(device, new System.Drawing.Font(
				"Arial", fontSize, System.Drawing.FontStyle.Bold | 
				System.Drawing.FontStyle.Italic));
			playerFontSprite = new Sprite(device);

			// Hook the font's events
			device.DeviceLost += new EventHandler(OnDeviceLost);
			device.DeviceReset += new EventHandler(OnDeviceReset);
			
			ai_state = AI_STATE.PATROL;

		}

		public AI(Device device, GameEngine parent, IntPtr parentHandle, string name, bool isLocal, string MeshFile, System.Drawing.Color myColor)
		{
			// Create a new Ship
			myShip = new Ship(device, MeshFile, myColor);

			// Store the player's name
			playerName = name;
			float fontSize = 36.0f;

			// Create the font for the player
			playerFont = new Font(device, new System.Drawing.Font(
				"Arial", fontSize, System.Drawing.FontStyle.Bold | 
				System.Drawing.FontStyle.Italic));
			playerFontSprite = new Sprite(device);

			// Hook the font's events
			device.DeviceLost += new EventHandler(OnDeviceLost);
			device.DeviceReset += new EventHandler(OnDeviceReset);
			
			ai_state = AI_STATE.PATROL;

		}

		#endregion

		#region IMoveableObject Members
		/// <summary>
		/// Draw the player (the Ship) in the world
		/// </summary>
		/// <param name="device">Device used to render Ship</param>
		public void Draw(GameEngine engine, Camera c)
		{
			if (myShip != null)
			{
				// Only draw the Ship if it lies in the view frustum
				if (c.ObjectInFrustum(this))
				{// Calculate where to put the player's ship
					Vector3 pos = Position;
					            
					// Set the world transform to the correct location
					engine.SetWorldTransform(myShip.ShipRotationMatrix 
						* Matrix.Translation(pos)
						* Matrix.Translation(myShip.Center));
	
					myShip.Draw(engine);					
				}
			}
		}
		/// <summary>
		/// Update the Ship in the world
		/// </summary>
		/// <param name="elapsedTime">Time elapsed since the last frame</param>
		public void Update(float elapsedTime)
		{
			System.Diagnostics.Debug.Assert(myShip != null, "Ship object cannot be null.");

			// Keep track of the total time
			totalTime += elapsedTime;


			// Finally update the ship
			myShip.Update(elapsedTime);
		}

		public bool ScanForTarget(IMoveableObject obj)
		{
			// The scan test for this will be simpler, simply compare the distance
			Vector3 diff = this.Position - obj.Position;
			return diff.Length() < scanRange;
		}

		public void SetTarget (IMoveableObject obj)
		{
			target = obj;
		}

		public IMoveableObject GetTarget()
		{
			return target;
		}

		public bool Intercept (float elapsedTime)
		{

			// find the larger distance, x or y
			Vector3 diff = this.Position - target.Position;
			if (diff.Length() < fireRange)
			{
				//within firing range
				//turn ship towards the enemy
				//Vector3 chance = Vector3.Normalize(this.Position - target.Position);
				//this.myShip.ShipDirection = chance; 

				float bearing = GetBearing(this.Position, target.Position);
				myShip.ShipRotation = bearing;
				if (diff.Length() > approachRange)
				{
					Vector3 dir = myShip.ShipDirection;
					dir.Scale((elapsedTime * MaximumMovementSpeed)); 
					myShip.Position += dir;
				}
				else if(diff.Length() > 400)
				{
					Vector3 dir = myShip.ShipDirection;
					dir.Scale((elapsedTime * MaximumMovementSpeed/4)); 
					myShip.Position += dir;
				}

				RaiseFireEvent();			

				//System.Diagnostics.Debug.WriteLine("AI player set heading"); 
			
				return true;
			}
			else
			{
				//move ship towards the enemy
				float bearing = GetBearing(this.Position, target.Position);
				myShip.ShipRotation = bearing ;
				
				// Movement is happening, first get the direction the tank is facing
				Vector3 dir = myShip.ShipDirection;                   
				dir.Scale((elapsedTime * MaximumMovementSpeed)); 
				//System.Diagnostics.Debug.WriteLine(dir.ToString());
				// Now append that to the current position
				//dir.X = Math.Abs(myShip.Position.X - dir.X);
				//dir.Z = Math.Abs(myShip.Position.Z - dir.Z);
				myShip.Position += dir;
				

                //System.Diagnostics.Debug.WriteLine("AI player set heading: "+ bearing.ToString() );                                                                                                                                                                                                                                                                                                                                                      
				return false;
			}
		}

		private float GetBearing( Vector3 self, Vector3 other )
		{
			float bearing = 0.0f;

			Vector3 direction = other - self;
			if ( Math.Abs(direction.X) > 0.1f ) //at the same X location
			{
				//System.Diagnostics.Debug.WriteLine("top bearing");
				if( direction.X >= 0 && direction.Z >= 0)
				{
					bearing = (float)Math.Atan(direction.X / direction.Z);
					//System.Diagnostics.Debug.WriteLine("top bearing1");
				}
				else if(( direction.X < 0 && direction.Z >= 0))
				{
					bearing = (float)Math.Atan(direction.X / direction.Z);
					//System.Diagnostics.Debug.WriteLine("top bearing2");
				}
				else if(( direction.X >= 0 && direction.Z < 0))
				{
					bearing = -(float)Math.PI+(float)Math.Atan(direction.X / direction.Z);
					//System.Diagnostics.Debug.WriteLine("top bearing3");
				}
				else if(( direction.X < 0 && direction.Z < 0))
				{
					bearing = (float)Math.PI+(float)Math.Atan(direction.X / direction.Z);
					//System.Diagnostics.Debug.WriteLine("top bearing4");
				}
			}
			else
			{
				//System.Diagnostics.Debug.WriteLine("BOTTOM bearing");
				if ( direction.Z > 0.0 )
				{
					bearing = 0.0f;
				}
				else
				{
					bearing = (float)Math.PI;
				}
			}
			return bearing;
		}

 
		#endregion

		#region Text Rendering
		/// <summary>
		/// Will render the player name above the tank
		/// </summary>
		public void DrawPlayerName(GameEngine engine)
		{
			// Calculate where to put the player's name
			Vector3 pos = Position;
			pos.Y += Radius / 1.75f;
            
			// Set the world transform to the correct location
			engine.SetWorldTransform(Matrix.RotationX((float)Math.PI) *
				myShip.ShipRotationMatrix * Matrix.Translation(pos)
				* Matrix.Translation(myShip.Center));

			// Now draw the text onto the sprite
			playerFontSprite.Begin( SpriteFlags.AlphaBlend | SpriteFlags.ObjectSpace);
			playerFont.DrawText(playerFontSprite, playerName, System.Drawing.Rectangle.Empty,
				DrawTextFormat.NoClip, System.Drawing.Color.CornflowerBlue);
			playerFontSprite.End();
            
		}
		#endregion

		#region GunFiring
		/// <summary>
		/// Get the starting bullet location and velocity
		/// </summary>
		public void GetBulletStartLocation(out Vector3 bulletStart, out Vector3 bulletVelocity)
		{
			// First, get the bullet start
			
			bulletStart = myShip.Position;
            
			// Scale the bullet in the correct direction to the end of the ship
			bulletStart.Add(Vector3.Scale(myShip.ShipDirection, Ship.ShipLength));
			bulletVelocity = Vector3.Scale(myShip.ShipDirection, Bullet.BulletStrenth);
			Vector3 deltaV = new Vector3(random.Next(-bullError,bullError),0,random.Next(-bullError,bullError));
			bulletVelocity += deltaV;
		}
		/// <summary>
		/// Raise the fire event of the gun if the cool down time has passed
		/// </summary>
		public bool RaiseFireEvent()
		{
			// Can the gun be fired again?
			if ((totalTime - lastFiredTime ) >= GunCoolDown)
			{
				// Yes it can, do so now
				if (Fired != null)
				{
					System.Diagnostics.Debug.WriteLine("AI fired");
					Fired(this, EventArgs.Empty);
				}
				// Store the last time the gun was fired
				lastFiredTime = totalTime;

				// The gun was fired
				System.Diagnostics.Debug.WriteLine("AI Raised Fired Event true"+totalTime.ToString()+" "+lastFiredTime.ToString());
				return true;
			}
			//System.Diagnostics.Debug.WriteLine("AI Raised Fired Event false. "+totalTime.ToString()+" "+lastFiredTime.ToString());
			return false;
		}

		/// <summary>
		/// Will add a new bullet to the list
		/// </summary>
		public void FireRound(Device device, AIBullets list)
		{
			// Get the bullet start and velocity
			Vector3 bulletStart;
			Vector3 bulletVel;
			GetBulletStartLocation(out bulletStart, out bulletVel);
			System.Diagnostics.Debug.WriteLine("AI Raised Fired with velocity of :   "+ bulletVel.ToString());

			// Finally add the new bullet
			list.Add(new AIBullet(device, bulletStart, bulletVel, this));

		}
		#endregion

             


       
		#region Device event handlers
		/// <summary>
		/// Fired right before the device is reset and about to be lost.
		/// </summary>
		private void OnDeviceLost(object sender, EventArgs e)
		{
			if ( (playerFont != null) && (!playerFont.Disposed) )
			{
				playerFont.OnLostDevice();
				playerFontSprite.OnLostDevice();
			}
		}
		/// <summary>
		/// Fired when a device is reset
		/// </summary>
		private void OnDeviceReset(object sender, EventArgs e)
		{
			if (playerFont != null)
			{
				playerFont.OnResetDevice();
				playerFontSprite.OnResetDevice();
			}
		}
		#endregion
 
		#region IDisposable Members
		/// <summary>
		/// Cleanup the player here
		/// </summary>
		public void Dispose()
		{
			// Cleanup the player stuff
			if (playerFontSprite != null)
			{
				playerFontSprite.Dispose();
				playerFont.Dispose();
			}

			// Cleanup the ship
			if (myShip != null)
			{
				myShip.Dispose();
			}

		}

		#endregion
	}

	/// <summary>
	/// A collection of AI ships
	/// </summary>
	public class AIShips : IEnumerable
	{
		#region Instance Data
		private ArrayList aiList = new ArrayList();
		#endregion

		#region Collection Methods
		/// <summary>
		/// Indexer for players
		/// </summary>
		public AI this[int index]
		{
			get { return aiList[index] as AI; }
			set { aiList[index] = value; }
		}

		/// <summary>
		/// Add a new player to the list
		/// </summary>
		public void Add(AI b)
		{
			aiList.Add(b);
		}
		/// <summary>
		/// Remove a player to the list
		/// </summary>
		public void Remove(AI b)
		{
			aiList.Remove(b);
		}

		/// <summary>
		/// Remove all players from the list
		/// </summary>
		public void Clear()
		{
			aiList.Clear();
		}

		/// <summary>
		/// The number of items in the list
		/// </summary>
		public int Count
		{
			get { return aiList.Count; }
		}
		#endregion

		#region IEnumerable Members
		/// <summary>
		/// Returns the enumerator for the player list
		/// </summary>
		public IEnumerator GetEnumerator()
		{
			return aiList.GetEnumerator();
		}
		#endregion
	}
}