using System;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace Demo
{
    /// <summary>
    /// This class will maintain the skybox used for the game
    /// </summary>
    public class Level : IDisposable
    {
        #region Private structure
        /// <summary>
        /// A structure used to define a single face
        /// </summary>
        private struct SingleFace
        {
            public Vector3[] positions;
            public Vector3 singleVertex;
            public Vector3 singleNormal;

            /// <summary>
            /// Return the plane that this face creates
            /// </summary>
            /// <returns></returns>
            public Plane GetFacePlane()
            {
                return Plane.FromPoints(positions[0], positions[1], positions[2]);
            }
            public Plane GetFacePlaneNormal()
            {
                return Plane.FromPointNormal(singleVertex, singleNormal);
            }
        }
        #endregion

        #region Constants
        private const float Height = 0.0f;
        private static readonly Matrix LevelMatrix = Matrix.Translation(0,Height,0);
        private const string LevelMeshName = "Level2.x";
        private const float LevelSize = 3000.0f;
        #endregion

        #region Instance Data
        private Mesh levelMesh = null;
        private Texture[] levelTextures = null;
        private SingleFace[] facePositions = null;
        #endregion

        #region Creation
        /// <summary>
        /// Create a new instance of the sky box class
        /// </summary>
        public Level(Device device)
        {
            // First create the mesh
            ExtendedMaterial[] mtrls = null;
            levelMesh = Mesh.FromFile(GameEngine.MediaPath + LevelMeshName, MeshFlags.Managed, device,
                out mtrls);

            // With the mesh created, create each of the textures used for the skybox
            if ((mtrls != null) && (mtrls.Length > 0))
            {
                levelTextures = new Texture[mtrls.Length];
                for(int i = 0; i < mtrls.Length; i++)
                {
                    levelTextures[i] = TexturePool.CreateTexture(device, mtrls[i].TextureFilename);
                }
            }

            // Allocate the correct number of faces
            facePositions = new SingleFace[levelMesh.NumberFaces];
            using (IndexBuffer ib = levelMesh.IndexBuffer)
            {
                // Find each face, first by locking the data
                short[] faces = ib.Lock(0, typeof(short), LockFlags.None, 
                    levelMesh.NumberFaces * 3) as short[];

                using (VertexBuffer vb = levelMesh.VertexBuffer)
                {
                    // Get a list of triangles that make the walls and floors
                    CustomVertex.PositionNormalTextured[] verts = vb.Lock(
                        0, typeof(CustomVertex.PositionNormalTextured), 
                        LockFlags.None, levelMesh.NumberVertices) as
                        CustomVertex.PositionNormalTextured[];

                    // Now, find each face vertex position
                    int faceIndex = -1;
                    for(int i = 0; i < faces.Length; i++)
                    {
                        if ( (i % 3) == 0)
                        {
                            // Allocate the position array
                            facePositions[++faceIndex].positions = new Vector3[3]; 
                        }
                        facePositions[faceIndex].positions[i % 3] = verts[faces[i]].Position;
                        facePositions[faceIndex].singleVertex = verts[faces[i]].Position;
                        facePositions[faceIndex].singleNormal = verts[faces[i]].Normal;
                    }
                    // Make sure to unlock the buffer
                    vb.Unlock();
                }

                // Make sure you unlock the index buffer
                ib.Unlock();
            }
        }
        #endregion

        #region Hit Testing
        /// <summary>
        /// Does a hit test on the object
        /// </summary>
        public bool HitTest(IMoveableObject obj, out Plane hitPlane)
        {
            // Default to an empty plane
            hitPlane = Plane.Empty;

            // This object hasn't left the level, has it hit a wall?
            IntersectInformation closestHit;
            if (levelMesh.Intersect(obj.Position, obj.Direction, out closestHit))
            {
                // This will almost always return true since you will always be hitting the outside of 
                // the level.  Now you need to check to see if you're close enough to really be hitting the 
                // closest point
                hitPlane = facePositions[closestHit.FaceIndex].GetFacePlane();
                return (closestHit.Dist < obj.Radius);
            }
            return false;
        }
        #endregion

        #region Rendering
        /// <summary>
        /// Render the level
        /// </summary>
        public void Draw(GameEngine engine, Device device)
        {
#if (DEBUG)
            engine.numberFaces += levelMesh.NumberFaces;
            engine.numberVerts += levelMesh.NumberVertices;
#endif
            // Set the world transform
            engine.SetWorldTransform(LevelMatrix);
            
            device.RenderState.Lighting = false;
            // Render each subset of the mesh
            for(int i = 0; i < levelTextures.Length; i++)
            {
                device.SetTexture(0, levelTextures[i]);
                levelMesh.DrawSubset(i);
            }
            device.RenderState.Lighting = true;
        }
        #endregion

        #region IDisposable Members
        /// <summary>
        /// Will be used to cleanup objects
        /// </summary>
        public void Dispose()
        {
            if (levelMesh != null)
            {
                levelMesh.Dispose();
                levelMesh = null;
            }
        }

        #endregion
    }
}