using System;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.DirectPlay;

namespace Demo
{
    /// <summary>
    /// Stores each possible network message (with a byte backend)
    /// </summary>
    public enum NetworkMessage : byte
    {
        UpdatePlayerLocation,
        UpdateshipTexture,
        PlayerFired,
    }

    /// <summary>
    /// This class will maintain the network engine for this game
    /// </summary>
    public class NetworkEngine : IDisposable
    {
        #region Constant Data
        private unsafe static readonly int PacketSizePlayerLocation = 
            sizeof(byte) + sizeof(Vector3) + sizeof(float);
        private unsafe static readonly int PacketSizePlayerTurret = 
            sizeof(byte) + sizeof(float) + sizeof(float);
        private unsafe static readonly int PacketSizePlayerFire = 
            sizeof(byte);
        private unsafe static readonly int PacketSizeColorChange = 
            sizeof(byte) + sizeof(byte);

        private const string SessionName = "'s game.";
        private static readonly Guid gameGuid = new Guid("{4695EAA3-B9A6-4388-A2FC-D2A36BA9A235}");
        private const int GamePort = 9897;

        private const int MaximumMessagesInQueue = 5;
        #endregion

        #region Instance Data
        private Peer networkDevice = null; // Will be used to maintain network state
        private Address deviceAddress = null; // Address for the device being used
        private bool isConnected = false; // Are you connected yet?
        private Player remotePlayer = null; // The remote player
        private int remotePlayerId = 0; // The remote player id
        #endregion

        #region Events
        public event PlayerCreatedEventHandler NewPlayer;
        public event PlayerDestroyedEventHandler DestroyedPlayer;
        #endregion

        #region Creation
        /// <summary>
        /// Constructor where the device will be created.
        /// </summary>
        public NetworkEngine()
        {
            // Create the peer object
            #if (DEBUG)
            networkDevice = new Peer();
            #else
            networkDevice = new Peer(InitializeFlags.DisableParameterValidation);
            #endif

            // Check if the internet provider is available
            if(!IsInternetAvailable())
            {
                throw new InvalidOperationException
                    ("You must have a valid TCP/IP connection to play this game.");
            }
            // Hook the events that we will care about
            networkDevice.Receive += new ReceiveEventHandler(OnDataReceive);
            networkDevice.ConnectComplete += new ConnectCompleteEventHandler(OnConnectComplete);
            networkDevice.PlayerCreated += new PlayerCreatedEventHandler(OnPlayerCreated);
            networkDevice.PlayerDestroyed += new PlayerDestroyedEventHandler(OnPlayerDestroyed);
            networkDevice.SessionTerminated += new SessionTerminatedEventHandler(OnSessionLost);
            networkDevice.FindHostResponse += new FindHostResponseEventHandler(OnFoundHost);

            // Create any addresses
            deviceAddress = new Address();
            // Use the TCP/IP provider on our port
            deviceAddress.ServiceProvider = Address.ServiceProviderTcpIp;
            deviceAddress.AddComponent(Address.KeyPort, GamePort);
        }
        #endregion

        #region Enter Session (Host/Join)
        /// <summary>
        /// Will host a session on your computer
        /// </summary>
        public void Host()
        {
            // Update with your information
            PlayerInformation info = new PlayerInformation();
            info.Name = Environment.UserName;
            networkDevice.SetPeerInformation(info, SyncFlags.PeerInformation);

            // Create the application description
            ApplicationDescription desc = new ApplicationDescription();
            desc.SessionName = Environment.UserName + SessionName; // Should be a relatively unique name
            desc.MaxPlayers = 2; // 1 on 1 battles
            desc.GuidApplication = gameGuid; // Unique id all games will use
            desc.Flags = SessionFlags.NoDpnServer; // We don't want to use the server

            // Now host the session
            networkDevice.Host(desc, deviceAddress, SessionName);
        }
        /// <summary>
        /// Join an existing session with the following host name
        /// </summary>
        public void Join(string hostName)
        {
            // Create the application description
            ApplicationDescription desc = new ApplicationDescription();
            desc.GuidApplication = gameGuid; // Unique id all games will use

            // Now try to find the host and connect to it
            using(Address localAddress = new Address())
            {
                // Set the local provider
                localAddress.ServiceProvider = Address.ServiceProviderTcpIp;
                // Set the hosts address name
                deviceAddress.AddComponent(Address.KeyHostname, hostName);

                // Search for hosts.
                networkDevice.FindHosts(desc, deviceAddress, localAddress, null, 0, 0, 0,
                    FindHostsFlags.None);
            }
        }
        #endregion

        #region IDisposable Members

        /// <summary>
        /// Done with this object, clean it up
        /// </summary>
        public void Dispose()
        {
            // Cleanup all data here
            if (networkDevice != null)
            {
                networkDevice.Dispose();
            }
            if (deviceAddress != null)
            {
                deviceAddress.Dispose();
            }
        }

        #endregion

        #region ServiceProvider check
        /// <summary>
        /// Checks to see if the TCP/IP Service Provider is available
        /// </summary>
        /// <returns>true if available; false otherwise</returns>
        private bool IsInternetAvailable()
        {
            // Ask about the providers attached
            ServiceProviderInformation[] providers = networkDevice.GetServiceProviders(
                false);

            // See if we have TCP/IP
            foreach(ServiceProviderInformation info in providers)
            {
                if (info.Guid == Address.ServiceProviderTcpIp)
                {
                    // we do, return true
                    return true;
                }
            }

            // We've made it here, the internet provider must not be available
            return false;
        }
        #endregion

        #region DirectPlay Event Handlers
        /// <summary>
        /// Fired when data is received from the other peer in the session
        /// </summary>
        private void OnDataReceive(object sender, ReceiveEventArgs e)
        {
            System.Diagnostics.Debug.Assert(remotePlayer != null, "Remote player should not be null.");
            using(NetworkPacket data = e.Message.ReceiveData)
            {
                NetworkMessage msg = (NetworkMessage)data.Read(typeof(byte));
                switch(msg)
                {
                    case NetworkMessage.UpdatePlayerLocation:
                        remotePlayer.Position = (Vector3)data.Read(typeof(Vector3));
                        remotePlayer.Rotation = (float)data.Read(typeof(float));
                        break;

                    case NetworkMessage.UpdateshipTexture:
                        //remotePlayer.Color = (ShipColors)data.Read(typeof(byte));
                        break;
                    case NetworkMessage.PlayerFired:
                        remotePlayer.RaiseFireEvent();
                        break;
                }
            }
        }
        /// <summary>
        /// Fired when the connection is completed
        /// </summary>
        private void OnConnectComplete(object sender, ConnectCompleteEventArgs e)
        {
            // Was the connection completed successfully?
            if (e.Message.ResultCode == ResultCode.Success)
            {
                isConnected = true;
                if (remotePlayer != null)
                {
                    //SendNewColor(remotePlayer.Color);
                }
            }
            else
            {
                // Well that sucks, the connection wasn't made
                System.Diagnostics.Debugger.Log(9, GameEngine.GameName,
                    string.Format("The connection failed.  Result was {0}.", 
                    e.Message.ResultCode));

                // Let the game engine know
                isConnected = false;
            }
        }
        /// <summary>
        /// Fired when a new player was created, including yourself
        /// </summary>
        private void OnPlayerCreated(object sender, PlayerCreatedEventArgs e)
        {
            // See if this player is the host
            if (e.Message.PlayerContext as string != SessionName)
            {
                // Store the player id
                remotePlayerId = e.Message.PlayerID;
                // This player is not, notify the game engine
                if (NewPlayer != null)
                {
                    NewPlayer(this, e);
                }

                // Update the playername too
                PlayerInformation info = networkDevice.GetPeerInformation(
                    e.Message.PlayerID);

                remotePlayer.Name = info.Name;
            }
        }
        /// <summary>
        /// Fired when a new player was destroyed, including yourself
        /// </summary>
        private void OnPlayerDestroyed(object sender, PlayerDestroyedEventArgs e)
        {
            // See if this player is the host
            if (e.Message.PlayerContext as string != SessionName)
            {
                // Remove the player id
                remotePlayerId = 0;
                // This player is not, notify the game engine
                if (DestroyedPlayer != null)
                {
                    DestroyedPlayer(this, e);
                }
            }
        }
        /// <summary>
        /// Fired when the networking session was lost
        /// </summary>
        private void OnSessionLost(object sender, SessionTerminatedEventArgs e)
        {
            // No longer connected.
            isConnected = false;
        }
        /// <summary>
        /// Fired when a host has been found
        /// </summary>
        private void OnFoundHost(object sender, FindHostResponseEventArgs e)
        {
            // If you've already connected, don't try again
            if (isConnected)
            {
                return;
            }

            // Try to connect now
            isConnected = true;
            Peer connection = sender as Peer;
            connection.Connect(e.Message.ApplicationDescription, e.Message.AddressSender,
                e.Message.AddressDevice, null, 0);
        }
        #endregion

        #region Send Data Methods
        /// <summary>
        /// Uses the send queue to determine if data can be sent
        /// </summary>
        /// <returns></returns>
        private bool CanSendData()
        {
            int numMessages = 0;
            int numBytes = 0;

            // Get the number of messages
            networkDevice.GetSendQueueInformation(remotePlayerId, out numMessages, out numBytes, 
                GetSendQueueInformationFlags.PriorityNormal);

            // Don't send if the message queue has more than 3 members
            return (numMessages < MaximumMessagesInQueue);
        }
        /// <summary>
        /// Will send the player's position and orientation to the other peer
        /// </summary>
        public bool SendPlayerLocation(Vector3 position, float rotation)
        {
            // Check to see if we can send this message
            if (CanSendData())
            {

                // Create a new network packet to send the player location
                NetworkPacket packet = new NetworkPacket(PacketSizePlayerLocation);
                // Write the message
                packet.Write((byte)NetworkMessage.UpdatePlayerLocation);
                packet.Write(position);
                packet.Write(rotation);
                // Now send the message
                networkDevice.SendTo(0, packet, 0, SendFlags.NoLoopback | SendFlags.NonSequential);
                // Message sent
                return true;
            }
            // The data cannot be sent, too many messages in queue
            return false;
        }
        
        /// <summary>
        /// Will send the player's position and orientation to the other peer
        /// </summary>
        public bool SendFireMessage()
        {
            // Create a new network packet to send the player location
            NetworkPacket packet = new NetworkPacket(PacketSizePlayerFire);
            // Write the message
            packet.Write((byte)NetworkMessage.PlayerFired);
            // Now send the message
            networkDevice.SendTo(0, packet, 0, SendFlags.NoLoopback | SendFlags.Guaranteed);
            // Since this is guaranteed, we should always send it, but let the game engine to slow 
            // down data transfer
            return CanSendData();
        }
        /// <summary>
        /// Will send the player's position and orientation to the other peer
        /// </summary>
        public bool SendNewColor(ShipColors newcolor)
        {
            // Create a new network packet to send the player location
            NetworkPacket packet = new NetworkPacket(PacketSizeColorChange);
            // Write the message
            packet.Write((byte)NetworkMessage.UpdateshipTexture);
            packet.Write((byte)newcolor);
            // Now send the message
            networkDevice.SendTo(0, packet, 0, SendFlags.NoLoopback | SendFlags.Guaranteed);
            // Since this is guaranteed, we should always send it, but let the game engine to slow 
            // down data transfer
            return CanSendData();
        }
        #endregion

        #region Remote player
        /// <summary>
        /// Set the remote player the network engine is controlling
        /// </summary>
        public void UpdatePlayer(Player p)
        {
            remotePlayer = p;
        }
        #endregion
    }
}