using System;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Input=Microsoft.DirectX.DirectInput;

namespace Tankers
{
    /// <summary>
    /// This class will maintain a player in the game
    /// </summary>
    public class Player : IDisposable, IMoveableObject
    {
        #region Constants
        private const float MaximumMovementSpeed = 600.0f;
        private const float MaximumMovementSpeedWheels = MaximumMovementSpeed / 100.0f;
        private const float MaximumRotationSpeed = 1.15f;
        private const float MaximumGunSpeed = 0.003f;
        private const float GunCoolDown = 0.2f;
        #endregion

        #region Instance Data
        private Ship gameTank = null;
        private bool isMovingForward = false;
        private bool isMovingBackward = false;
        private bool isRotatingLeft = false;
        private bool isRotatingRight = false;
        private bool isPlayerLocal = true;
        private bool isPlayerActive = false; // Only for remote players
        
        // Default the mouse stats to -1 so we know they haven't been used yet.
        private int mouseX = -1;
        private int mouseY = -1;

        // The device for DirectInput, will only be used for the local player
        private Input.Device joystickDevice = null;
        private bool hasHat = false; // Will be used to determine if the joystick has a hat for control
        private bool isMovingForwardJoy = false;
        private bool isMovingBackwardJoy = false;
        private bool isRotatingLeftJoy = false;
        private bool isRotatingRightJoy = false;

        // Tank firing items
        public event EventHandler Fired;
        private float lastFiredTime = -GunCoolDown; // So you can fire right away
        private float totalTime = 0.0f;

        // Store the players name and allow it to be rendered
        private string playerName = string.Empty;
        private Font playerFont = null;
        private Sprite playerFontSprite = null;
        #endregion

        #region Creation
        /// <summary>
        /// Create a new player instance
        /// </summary>
        /// <param name="device">Device used for rendering</param>
        /// <param name="parent">Parent control used to hook events</param>
        public Player(Device device, GameEngine parent, IntPtr parentHandle, string name, bool isLocal)
        {
            // Create a new tank
            gameTank = new Ship(device);

            // Store the info here
            isPlayerLocal = isLocal;

            // Hook the events if the player is local
            if (isPlayerLocal)
            {
                //Hook windows events for user input
                parent.KeyDown += new KeyEventHandler(OnKeyDown);
                parent.KeyUp += new KeyEventHandler(OnKeyUp);
                // Hook mouse events too
                parent.MouseMove += new MouseEventHandler(OnMouseMove);
                parent.MouseClick += new MouseEventHandler(OnMouseClick);

                InitializeJoystick(parentHandle);
            }

            // Store the player's name
            playerName = name;
            float fontSize = isPlayerLocal ? 12.0f : 36.0f;

            // Create the font for the player
            playerFont = new Font(device, new System.Drawing.Font(
                "Arial", fontSize, System.Drawing.FontStyle.Bold | 
                System.Drawing.FontStyle.Italic));
            playerFontSprite = new Sprite(device);

            // Hook the font's events
            device.DeviceLost += new EventHandler(OnDeviceLost);
            device.DeviceReset += new EventHandler(OnDeviceReset);
        }

        /// <summary>
        /// Will initialize the joystick if present for this user
        /// </summary>
        private void InitializeJoystick(IntPtr parentHandle)
        {
            // See if there is a joystick attached to the system
            // Enumerate joysticks in the system.
            foreach (Input.DeviceInstance instance in Input.Manager.GetDevices(
                Input.DeviceClass.GameControl, Input.EnumDevicesFlags.AttachedOnly))
            {
                // Create the device.  Just pick the first one
                joystickDevice = new Input.Device(instance.InstanceGuid);
                break;
            }

            bool hasButton = false;
            bool hasXAxis = false;
            bool hasYAxis = false;

            // If no joystick was found, simply exit this method after a quick message
            if (joystickDevice == null)
            {
                System.Diagnostics.Debugger.Log(0, GameEngine.GameName + ": Warning"
                    , "No joystick found, will use mouse and keyboard");
            }
            else
            {
                // There is a joystick attached, update some state for the joystick

                // Set the data format to the pre-defined format.
                joystickDevice.SetDataFormat(Input.DeviceDataFormat.Joystick);
                // Set the cooperative level for the device.
                joystickDevice.SetCooperativeLevel(Form.FromHandle(parentHandle), Input.CooperativeLevelFlags.NonExclusive 
                    | Input.CooperativeLevelFlags.Background);

                // Enumerate all the objects on the device.
                foreach (Input.DeviceObjectInstance d in joystickDevice.Objects)
                {
                    if ((d.ObjectId & (int)Input.DeviceObjectTypeFlags.Axis) != 0)
                    {
                        // This is an axis, set it's range
                        joystickDevice.Properties.SetRange(
                            Input.ParameterHow.ById, d.ObjectId,
                            new Input.InputRange(-5000, +5000));

                        // Set the dampening zone of this as well
                        joystickDevice.Properties.SetDeadZone(
                            Input.ParameterHow.ById, d.ObjectId,
                            500);
                    }

                    // See if we have a hat that will be used for gun turret movement
                    if (d.ObjectType == Input.ObjectTypeGuid.PointOfView)
                    {
                        hasHat = true;
                    }

                    // Make sure we have at least one button and an x/y axis
                    if (d.ObjectType == Input.ObjectTypeGuid.Button)
                    {
                        hasButton = true;
                    }
                    if (d.ObjectType == Input.ObjectTypeGuid.XAxis)
                    {
                        hasXAxis = true;
                    }
                    if (d.ObjectType == Input.ObjectTypeGuid.YAxis)
                    {
                        hasYAxis = true;
                    }
                }
                // Now if there isn't an x/y axis and button, reset the joystick to null since it can't be used.
                if ( (!hasButton) || (!hasXAxis) || (!hasYAxis) )
                {
                    System.Diagnostics.Debugger.Log(0, GameEngine.GameName + ": Warning"
                        , "Joystick found but without necessary features, will use mouse and keyboard");

                    joystickDevice.Dispose();
                    joystickDevice = null;
                }
            }
        }
        #endregion

        #region IMoveableObject Members
        /// <summary>
        /// Draw the player (the tank) in the world
        /// </summary>
        /// <param name="device">Device used to render tank</param>
        public void Draw(GameEngine engine, Camera c)
        {
            if (gameTank != null)
            {
                // Only draw the tank if it lies in the view frustum
                if (c.ObjectInFrustum(this))
                {
                    gameTank.Draw(engine);
                }
            }
        }
        /// <summary>
        /// Update the tank in the world
        /// </summary>
        /// <param name="elapsedTime">Time elapsed since the last frame</param>
        public void Update(float elapsedTime, Level currentLevel)
        {
            System.Diagnostics.Debug.Assert(gameTank != null, "Tank object cannot be null.");

            // Keep track of the total time
            totalTime += elapsedTime;

            // Update movement if the player is local
            if (isPlayerLocal)
            {
                // If the joystick can be used, try to use it
                if (joystickDevice != null)
                {
                    UpdateJoystick(joystickDevice);
                }

                // Is the player moving?
                if ((isMovingForward) || (isMovingBackward) || (isMovingForwardJoy) || (isMovingBackwardJoy))
                {
                    // Movement is happening, first get the direction the tank is facing
                    Vector3 dir = gameTank.TankDirection;

                    // Now scale it to move, first check keyboard
                    if ((isMovingForward) || (isMovingBackward))
                    {
                        dir.Scale((isMovingForward) ? (elapsedTime * MaximumMovementSpeed) : 
                            (elapsedTime * -MaximumMovementSpeed));
                    }
                    else // Must be joystick
                    {
                        dir.Scale((isMovingForwardJoy) ? (elapsedTime * MaximumMovementSpeed) : 
                            (elapsedTime * -MaximumMovementSpeed));
                    }

                    // Now append that to the current position
                    gameTank.Position += dir;

                    Plane reflectionPlane;
                    if (currentLevel.HitTest(this, out reflectionPlane))
                    {
                        // Well, you're hitting a wall.  Backup
                        gameTank.Position -= dir;
                        // Now stop moving
                        isMovingBackward = false;
                        isMovingForward = false;
                    }
                    // Now update the wheel rotation animation
                    gameTank.WheelRotation += (isMovingForward) ? (elapsedTime * -MaximumMovementSpeedWheels) : 
                        (elapsedTime * MaximumMovementSpeedWheels);
                }
                // Is the player rotating?
                if ((isRotatingLeft) || (isRotatingRight) || (isRotatingLeftJoy) || (isRotatingRightJoy))
                {
                    // First check keyboard
                    if ((isRotatingLeft) || (isRotatingRight))
                    {
                        gameTank.TankRotation += (isRotatingRight) ? (MaximumRotationSpeed * elapsedTime) : 
                            (-MaximumRotationSpeed * elapsedTime);
                    }
                    else // Must be joystick
                    {
                        gameTank.TankRotation += (isRotatingRightJoy) ? (MaximumRotationSpeed * elapsedTime) : 
                            (-MaximumRotationSpeed * elapsedTime);
                    }
                }
            }
            // Finally update the tank
            gameTank.Update(elapsedTime);
        }

        /// <summary>
        /// The tanks current position
        /// </summary>
        public Vector3 Position
        {
            get { return gameTank.Position; }
            set { gameTank.Position = value; }
        }
        /// <summary>
        /// The direction the player is currently facing
        /// </summary>
        public Vector3 Direction
        {
            get { return gameTank.TankDirection; }
        }

        /// <summary>
        /// The current rotation of the tank
        /// </summary>
        public float Rotation
        {
            get { return gameTank.TankRotation; }
            set { gameTank.TankRotation = value; }
        }
        /// <summary>
        /// Radius of the player model
        /// </summary>
        public float Radius
        {
            get { return gameTank.Radius; }
        }
        #endregion

        #region Text Rendering
        /// <summary>
        /// Will render the player name above the tank
        /// </summary>
        public void DrawPlayerName(GameEngine engine)
        {
            // Calculate where to put the player's name
            Vector3 pos = Position;
            pos.Y += Radius / 1.75f;
            
            // Set the world transform to the correct location
            engine.SetWorldTransform(Matrix.RotationX((float)Math.PI) *
                gameTank.TankRotationMatrix * Matrix.Translation(pos)
                * Matrix.Translation(gameTank.Center));

            // Now draw the text onto the sprite
            playerFontSprite.Begin( SpriteFlags.AlphaBlend | SpriteFlags.ObjectSpace);
            playerFont.DrawText(playerFontSprite, playerName, System.Drawing.Rectangle.Empty,
                DrawTextFormat.NoClip, System.Drawing.Color.CornflowerBlue);
            playerFontSprite.End();
            
        }
        #endregion

        #region GunFiring
        /// <summary>
        /// Get the starting bullet location and velocity
        /// </summary>
        public void GetBulletStartLocation(out Vector3 bulletStart, out Vector3 bulletVelocity)
        {
            // First, get the bullet start
            bulletStart = gameTank.BarrelPosition;
            
            // Scale the bullet in the correct direction to the end of the barrel
            bulletStart.Add(Vector3.Scale(gameTank.BarrelDirection, Tank.GunBarrelLength));
            bulletVelocity = Vector3.Scale(gameTank.BarrelDirection, Bullet.BulletStrenth);
        }
        /// <summary>
        /// Raise the fire event of the gun if the cool down time has passed
        /// </summary>
        public bool RaiseFireEvent()
        {
            // Can the gun be fired again?
            if ((totalTime - lastFiredTime ) > GunCoolDown)
            {
                // Yes it can, do so now
                if (Fired != null)
                {
                    Fired(this, EventArgs.Empty);
                }
                // Store the last time the gun was fired
                lastFiredTime = totalTime;

                // The gun was fired
                return true;
            }
            return false;
        }

        /// <summary>
        /// Will add a new bullet to the list
        /// </summary>
        public void FireRound(Device device, Bullets list)
        {
            // Get the bullet start and velocity
            Vector3 bulletStart;
            Vector3 bulletVel;
            GetBulletStartLocation(out bulletStart, out bulletVel);

            // Finally add the new bullet
            list.Add(new Bullet(device, bulletStart, bulletVel, this));
        }
        #endregion

        #region Windows Forms Events
        /// <summary>
        /// Will handle keyboard events for the player
        /// </summary>
        private void OnKeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.W)
            {
                isMovingForward = true;
            }
            else if (e.KeyCode == Keys.S)
            {
                isMovingBackward = true;
            }
            else if (e.KeyCode == Keys.A)
            {
                isRotatingLeft = true;
            }
            else if (e.KeyCode == Keys.D)
            {
                isRotatingRight = true;
            }
        }
        /// <summary>
        /// Will handle keyboard events for the player
        /// </summary>
        private void OnKeyUp(object sender, KeyEventArgs e)
        {
            // depending on which key was let up, reset the variables
            if (e.KeyCode == Keys.W)
            {
                isMovingForward = false;
            }
            else if (e.KeyCode == Keys.S)
            {
                isMovingBackward = false;
            }
            else if (e.KeyCode == Keys.A)
            {
                isRotatingLeft = false;
            }
            else if (e.KeyCode == Keys.D)
            {
                isRotatingRight = false;
            }
        }
        /// <summary>
        /// Will handle mouse moving events for the player
        /// </summary>
        private void OnMouseMove(object sender, MouseEventArgs e)
        {
            if ((mouseX == -1) || (mouseY == -1))
            {
                // set the first mouse coordinates
                mouseX = e.X;
                mouseY = e.Y;
                // Nothing else to do now
                return;
            }

            if (gameTank != null)
            {
                gameTank.GunTurretAngle += ((e.X - mouseX) * MaximumGunSpeed);
                gameTank.GunBarrelAngle += ((e.Y - mouseY) * MaximumGunSpeed);
            }

            // Save the mouse coordinates
            mouseX = e.X;
            mouseY = e.Y;
        }

        /// <summary>
        /// Used when the mouse button is release (clicked)
        /// </summary>
        private void OnMouseClick(object sender, MouseEventArgs e)
        {
            // If they clicked the mouse button, fire
            RaiseFireEvent();
        }
        #endregion

        #region UserInput(Joystick)
        /// <summary>
        /// Will update the player data based on this joystick device
        /// </summary>
        private void UpdateJoystick(Input.Device joystick)
        {
            try
            {
                // See if we can get the joystick data
                joystick.Poll();
            }
            catch(Input.InputException e)
            {
                if ((e is Input.NotAcquiredException) || (e is Input.InputLostException))
                {
                    // If the joystick wasn't acquired, or was lost, try to acquire it now
                    try
                    {
                        // Acquire the device.
                        joystick.Acquire();
                    }
                    catch(Input.InputException)
                    {
                        // Failed to acquire the device, nothing we can do, exit
                        return;
                    }
                }
            }

            // Get the state of the device.
            Input.JoystickState state = joystick.CurrentJoystickState;

            // Now update the player
            if (state.X > 0) { isRotatingRightJoy = true; isRotatingLeftJoy = false; }
            else if (state.X < 0) { isRotatingLeftJoy = true; isRotatingRightJoy = false;}
            else { isRotatingLeftJoy = false; isRotatingRightJoy = false;}
            if (state.Y < 0) { isMovingForwardJoy = true; isMovingBackwardJoy = false; }
            else if (state.Y > 0) { isMovingBackwardJoy = true; isMovingForwardJoy = false;}
            else {isMovingBackwardJoy = false; isMovingForwardJoy = false; }

            // See if the first button has been fired
            byte[] buttons = state.GetButtons();
            if ((buttons[0] & 0x80) != 0)
            {
                // Fire!
                RaiseFireEvent();
            }

            // Check the hat
            if (hasHat)
            {
                // Update the gun barrel based on POV
                int[] pov = state.GetPointOfView();
                switch(pov[0])
                {
                    case 0:
                        // Move barrel up
                        gameTank.GunBarrelAngle -= (MaximumGunSpeed);
                        break;
                    case 4500:
                        // Move barrel up, and turret over
                        gameTank.GunBarrelAngle -= (MaximumGunSpeed);
                        gameTank.GunTurretAngle += (MaximumGunSpeed);
                        break;
                    case 9000:
                        // Move turret over
                        gameTank.GunTurretAngle += (MaximumGunSpeed);
                        break;
                    case 13500:
                        // Move barrel down, and turret over
                        gameTank.GunBarrelAngle += (MaximumGunSpeed);
                        gameTank.GunTurretAngle += (MaximumGunSpeed);
                        break;
                    case 18000:
                        // Move barrel down
                        gameTank.GunBarrelAngle += (MaximumGunSpeed);
                        break;
                    case 22500:
                        // Move barrel down, and turret over
                        gameTank.GunBarrelAngle += (MaximumGunSpeed);
                        gameTank.GunTurretAngle -= (MaximumGunSpeed);
                        break;
                    case 27000:
                        // Move turret over
                        gameTank.GunTurretAngle -= (MaximumGunSpeed);
                        break;
                    case 31500:
                        // Move barrel up, and turret over
                        gameTank.GunBarrelAngle -= (MaximumGunSpeed);
                        gameTank.GunTurretAngle -= (MaximumGunSpeed);
                        break;
                }
            }
        }
        #endregion

        #region Network methods/properties
        /// <summary>
        /// Used to determine if the remote player is connected 
        /// </summary>
        public bool IsPlayerConnected
        {
            get { return isPlayerActive; }
            set { isPlayerActive = value; }
        }

        /// <summary>
        /// Is the player the local player
        /// </summary>
        public bool IsLocal
        {
            get { return isPlayerLocal; }
        }

        /// <summary>
        /// Simply pass on the color information here
        /// </summary>
        public TankColors Color
        {
            get { return gameTank.Color; }
            set { gameTank.Color = value; }
        }

        /// <summary>
        /// Updates the gun turret and barrel
        /// </summary>
        public void UpdateTurret(float turret, float gun)
        {
            gameTank.GunTurretAngle = turret;
            gameTank.GunBarrelAngle = gun;
        }
        public float GunTurretAngle
        {
            get { return gameTank.GunTurretAngle; }
        }
        public float GunBarrelAngle
        {
            get { return gameTank.GunBarrelAngle; }
        }

        /// <summary>
        /// Name of the player
        /// </summary>
        public string Name
        {
            get { return playerName; }
            set { playerName = value; }
        }
        #endregion

        #region Device event handlers
        /// <summary>
        /// Fired right before the device is reset and about to be lost.
        /// </summary>
        private void OnDeviceLost(object sender, EventArgs e)
        {
            if ( (playerFont != null) && (!playerFont.Disposed) )
            {
                playerFont.OnLostDevice();
                playerFontSprite.OnLostDevice();
            }
        }
        /// <summary>
        /// Fired when a device is reset
        /// </summary>
        private void OnDeviceReset(object sender, EventArgs e)
        {
            if (playerFont != null)
            {
                playerFont.OnResetDevice();
                playerFontSprite.OnResetDevice();
            }
        }
        #endregion
 
        #region IDisposable Members
        /// <summary>
        /// Cleanup the player here
        /// </summary>
        public void Dispose()
        {
            // Cleanup the player stuff
            if (playerFontSprite != null)
            {
                playerFontSprite.Dispose();
                playerFont.Dispose();
            }

            // Cleanup the tank
            if (gameTank != null)
            {
                gameTank.Dispose();
            }

            // Cleanup the joystick if it exists
            if (joystickDevice != null)
            {
                joystickDevice.Unacquire();
                joystickDevice.Dispose();
            }
            joystickDevice = null;
        }

        #endregion
    }
}