using System;
using System.Collections;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Direct3D=Microsoft.DirectX.Direct3D;
using Input=Microsoft.DirectX.DirectInput;

namespace Demo
{
    /// <summary>
    /// This class will maintain a player in the game
    /// </summary>
    public class Player : IDisposable, IMoveableObject
    {
        #region Constants
        private const float MaximumMovementSpeed = 600.0f;
        private const float MaximumRotationSpeed = 1.15f;
		private const float MaximumClimbSpeed = 1.15f;
        private const float GunCoolDown = 0.2f;
        #endregion

        #region Instance Data
        private Ship myShip = null;
        private bool isMovingForward = false;
        private bool isMovingBackward = false;
        private bool isRotatingLeft = false;
        private bool isRotatingRight = false;

        private bool isPlayerLocal = true;
        private bool isPlayerActive = false; // Only for remote players
		private string shipMesh = null;
        
        // Default the mouse stats to -1 so we know they haven't been used yet.
        private int mouseX = -1;
        private int mouseY = -1;
		private float MAX_MOVE = 60000f;
		private float MIN_MOVE = -60000f;

        // ship firing items
        public event EventHandler Fired;
        private float lastFiredTime = -GunCoolDown; // So you can fire right away
        private float totalTime = 0.0f;

        // Store the players name and allow it to be rendered
        private string playerName = string.Empty;
        private Microsoft.DirectX.Direct3D.Font playerFont = null;
        private Direct3D.Sprite playerFontSprite = null;
		private int armor = 100;
		private bool Death = false;
		

		public int Armor
		{
			get { return armor;}
			set { armor = value;}
		}

		public bool Dead
		{
			get {return Death;}
			set {Death = value;}
		}

			/// <summary>
			/// The ships current position
			/// </summary>
			public Vector3 Position
		{
			get { return myShip.Position; }
			set { myShip.Position = value; }
		}
		/// <summary>
		/// The direction the player is currently facing
		/// </summary>
		public Vector3 Direction
		{
			get { return myShip.ShipDirection; }
		}

		public float Orientation
		{
			get { return myShip.ShipOrientation; }
			set { myShip.ShipOrientation = value; }
		}

		/// <summary>
		/// The current rotation of the ship
		/// </summary>
		public float Rotation
		{
			get { return myShip.ShipRotation; }
			set { myShip.ShipRotation = value; }
		}
		/// <summary>
		/// Radius of the player model
		/// </summary>
		public float Radius
		{
			get { return myShip.Radius; }
		}
        #endregion

        #region Creation
        /// <summary>
        /// Create a new player instance
        /// </summary>
        /// <param name="device">Device used for rendering</param>
        /// <param name="parent">Parent control used to hook events</param>
        public Player(Direct3D.Device device, GameEngine parent, IntPtr parentHandle, string name, bool isLocal)
        {
            // Create a new ship
            myShip = new Ship(device);

            // Store the info here
            isPlayerLocal = isLocal;

            // Hook the events if the player is local
            if (isPlayerLocal)
            {
                //Hook windows events for user input
                parent.KeyDown += new KeyEventHandler(OnKeyDown);
                parent.KeyUp += new KeyEventHandler(OnKeyUp);
                // Hook mouse events too
                parent.MouseMove += new MouseEventHandler(OnMouseMove);
                parent.MouseClick += new MouseEventHandler(OnMouseClick);

            }

            // Store the player's name
            playerName = name;
            float fontSize = isPlayerLocal ? 12.0f : 36.0f;

            // Create the font for the player
            playerFont = new Microsoft.DirectX.Direct3D.Font(device, new System.Drawing.Font(
                "Arial", fontSize, System.Drawing.FontStyle.Bold | 
                System.Drawing.FontStyle.Italic));
            playerFontSprite = new Sprite(device);

            // Hook the font's events
            device.DeviceLost += new EventHandler(OnDeviceLost);
            device.DeviceReset += new EventHandler(OnDeviceReset);
        }

		public Player(Device device, GameEngine parent, IntPtr parentHandle, string name, bool isLocal, string MeshFile)
		{
			// Create a new ship
			myShip = new Ship(device, MeshFile);

			// Store the info here
			isPlayerLocal = isLocal;

			// Hook the events if the player is local
			if (isPlayerLocal)
			{
				//Hook windows events for user input
				parent.KeyDown += new KeyEventHandler(OnKeyDown);
				parent.KeyUp += new KeyEventHandler(OnKeyUp);
				// Hook mouse events too
				parent.MouseMove += new MouseEventHandler(OnMouseMove);
				parent.MouseClick += new MouseEventHandler(OnMouseClick);
				

			}

			// Store the player's name
			playerName = name;
			float fontSize = isPlayerLocal ? 12.0f : 36.0f;

			// Create the font for the player
			playerFont = new Direct3D.Font(device, new System.Drawing.Font(
				"Arial", fontSize, System.Drawing.FontStyle.Bold | 
				System.Drawing.FontStyle.Italic));
			playerFontSprite = new Sprite(device);

			// Hook the font's events
			device.DeviceLost += new EventHandler(OnDeviceLost);
			device.DeviceReset += new EventHandler(OnDeviceReset);
			
		}

		public Player(Device device, GameEngine parent, IntPtr parentHandle, string name, bool isLocal, string MeshFile, Color myColor)
		{
			// Create a new ship
			myShip = new Ship(device, MeshFile, myColor);
			shipMesh = MeshFile;
			

			// Store the info here
			isPlayerLocal = isLocal;

			// Hook the events if the player is local
			if (isPlayerLocal)
			{
				//Hook windows events for user input
				parent.KeyDown += new KeyEventHandler(OnKeyDown);
				parent.KeyUp += new KeyEventHandler(OnKeyUp);
				// Hook mouse events too
				parent.MouseMove += new MouseEventHandler(OnMouseMove);
				parent.MouseClick += new MouseEventHandler(OnMouseClick);
				

			}

			// Store the player's name
			playerName = name;
			float fontSize = isPlayerLocal ? 12.0f : 36.0f;

			// Create the font for the player
			playerFont = new Direct3D.Font(device, new System.Drawing.Font(
				"Arial", fontSize, System.Drawing.FontStyle.Bold | 
				System.Drawing.FontStyle.Italic));
			playerFontSprite = new Sprite(device);

			// Hook the font's events
			device.DeviceLost += new EventHandler(OnDeviceLost);
			device.DeviceReset += new EventHandler(OnDeviceReset);
			
		}
        #endregion

        #region IMoveableObject Members
        /// <summary>
        /// Draw the player (the ship) in the world
        /// </summary>
        /// <param name="device">Device used to render ship</param>
        public void Draw(GameEngine engine, Camera c)
        {
            if (myShip != null)
            {
                // Only draw the ship if it lies in the view frustum
                if (c.ObjectInFrustum(this))
                {// Calculate where to put the player's ship
										            
					// Set the world transform to the correct location
					if (shipMesh == "viper4.x")
					{
						engine.SetWorldTransform(Matrix.RotationY((float)Math.PI/2) * myShip.ShipRotationMatrix 
							* Matrix.Translation(Position)
							* Matrix.Translation(myShip.Center)
						
							);
					}
					else if (shipMesh == "raider.x")
					{
						engine.SetWorldTransform(myShip.ShipRotationMatrix 
							* Matrix.Translation(Position)
							* Matrix.Translation(myShip.Center)
						
							);
					}
	
                    myShip.Draw(engine);					
                }
            }
        }
        /// <summary>
        /// Update the ship in the world
        /// </summary>
        /// <param name="elapsedTime">Time elapsed since the last frame</param>
        public void Update(float elapsedTime)
        {
            System.Diagnostics.Debug.Assert(myShip != null, "ship object cannot be null.");

            // Keep track of the total time
            totalTime += elapsedTime;

            // Update movement if the player is local
            if (isPlayerLocal)
            {
                // Is the player moving?
                if ((isMovingForward) || (isMovingBackward))
                {
                    // Movement is happening, first get the direction the ship is facing
                    Vector3 dir = myShip.ShipDirection;                   
                   dir.Scale((isMovingForward) ? (elapsedTime * MaximumMovementSpeed) : 
                            (elapsedTime * -MaximumMovementSpeed)); 
					//System.Diagnostics.Debug.WriteLine(dir.ToString());
                    // Now append that to the current position
                    myShip.Position += dir;
					if ( myShip.Position.X > MAX_MOVE) myShip.Position = new Vector3(MAX_MOVE, myShip.Position.Y, myShip.Position.Z);
					if ( myShip.Position.Z > MAX_MOVE) myShip.Position = new Vector3(myShip.Position.X, myShip.Position.Y, MAX_MOVE);
					if ( myShip.Position.X < MIN_MOVE) myShip.Position = new Vector3(MIN_MOVE, myShip.Position.Y, myShip.Position.Z);
					if ( myShip.Position.Z < MIN_MOVE) myShip.Position = new Vector3(myShip.Position.X, myShip.Position.Y, MIN_MOVE);;
					//System.Diagnostics.Debug.WriteLine("final position "+myShip.Position.ToString());
                }
                // Is the player rotating?
                if ((isRotatingLeft) || (isRotatingRight))
                {
                        myShip.ShipRotation += (isRotatingRight) ? (MaximumRotationSpeed * elapsedTime) : 
                            (-MaximumRotationSpeed * elapsedTime);
                }
            }
            // Finally update the ship
            myShip.Update(elapsedTime);
        }


 
        #endregion

        #region Text Rendering
        /// <summary>
        /// Will render the player name above the ship
        /// </summary>
        public void DrawPlayerName(GameEngine engine)
        {
            // Calculate where to put the player's name
            Vector3 pos = Position;
            pos.Y += Radius / 1.75f;
            
            // Set the world transform to the correct location
            engine.SetWorldTransform(Matrix.RotationX((float)Math.PI) *
                myShip.ShipRotationMatrix * Matrix.Translation(pos)
                * Matrix.Translation(myShip.Center));

            // Now draw the text onto the sprite
            playerFontSprite.Begin( SpriteFlags.AlphaBlend | SpriteFlags.ObjectSpace);
            playerFont.DrawText(playerFontSprite, playerName, System.Drawing.Rectangle.Empty,
                DrawTextFormat.NoClip, System.Drawing.Color.CornflowerBlue);
            playerFontSprite.End();
            
        }
        #endregion

        #region GunFiring
        /// <summary>
        /// Get the starting bullet location and velocity
        /// </summary>
        public void GetBulletStartLocation(out Vector3 bulletStart, out Vector3 bulletVelocity)
        {
            // First, get the bullet start
           // bulletStart = myShip.ShipDirection;
			bulletStart = myShip.Position;
            
            // Scale the bullet in the correct direction to the end of the barrel
            bulletStart.Add(Vector3.Scale(myShip.ShipDirection, Ship.ShipLength));
            bulletVelocity = Vector3.Scale(myShip.ShipDirection, Bullet.BulletStrenth);
        }
        /// <summary>
        /// Raise the fire event of the gun if the cool down time has passed
        /// </summary>
        public bool RaiseFireEvent()
        {
            // Can the gun be fired again?
            if ((totalTime - lastFiredTime ) > GunCoolDown)
            {
                // Yes it can, do so now
                if (Fired != null)
                {
                    Fired(this, EventArgs.Empty);
                }
                // Store the last time the gun was fired
                lastFiredTime = totalTime;

                // The gun was fired
                return true;
            }
            return false;
        }

        /// <summary>
        /// Will add a new bullet to the list
        /// </summary>
        public void FireRound(Device device, Bullets list)
        {
            // Get the bullet start and velocity
            Vector3 bulletStart;
            Vector3 bulletVel;
            GetBulletStartLocation(out bulletStart, out bulletVel);

            // Finally add the new bullet
            list.Add(new Bullet(device, bulletStart, bulletVel, this));

        }
        #endregion

        #region Windows Forms Events
        /// <summary>
        /// Will handle keyboard events for the player
        /// </summary>
        private void OnKeyDown(object sender, KeyEventArgs e)
        {
			if (e.KeyCode == Keys.W)
			{
				isMovingForward = true;
			}
			else if (e.KeyCode == Keys.S)
			{
				isMovingBackward = true;
			}
			else if (e.KeyCode == Keys.A)
			{
				isRotatingLeft = true;
			}
			else if (e.KeyCode == Keys.D)
			{
				isRotatingRight = true;
			}
			else if (e.KeyCode == Keys.Left)
			{
				isRotatingLeft = true;
			}
			else if (e.KeyCode == Keys.Right)
			{
				isRotatingRight = true;
			}
			else if (e.KeyCode == Keys.Up)
			{
				isMovingForward = true;
			}
			else if (e.KeyCode == Keys.Down)
			{
				isMovingBackward = true;
			}
        }
        /// <summary>
        /// Will handle keyboard events for the player
        /// </summary>
        private void OnKeyUp(object sender, KeyEventArgs e)
        {
            // depending on which key was let up, reset the variables
			if (e.KeyCode == Keys.W)
			{
				isMovingForward = false;
			}
			else if (e.KeyCode == Keys.S)
			{
				isMovingBackward = false;
			}
			else if (e.KeyCode == Keys.A)
			{
				isRotatingLeft = false;
			}
			else if (e.KeyCode == Keys.D)
			{
				isRotatingRight = false;
			}
			else if (e.KeyCode == Keys.Left)
			{
				isRotatingLeft = false;
			}
			else if (e.KeyCode == Keys.Right)
			{
				isRotatingRight = false;
			}
			else if (e.KeyCode == Keys.Up)
			{
				isMovingForward = false;
			}
			else if (e.KeyCode == Keys.Down)
			{
				isMovingBackward = false;
			}
        }
        /// <summary>
        /// Will handle mouse moving events for the player
        /// </summary>
        private void OnMouseMove(object sender, MouseEventArgs e)
        {
            if ((mouseX == -1) || (mouseY == -1))
            {
                // set the first mouse coordinates
                mouseX = e.X;
                mouseY = e.Y;
                // Nothing else to do now
                return;
            }

            if (myShip != null)
            {
                myShip.ShipRotation += ((e.X - mouseX) * MaximumRotationSpeed / 100);
                myShip.ShipRotation += ((e.Y - mouseY) * MaximumRotationSpeed / 100);
            }

            // Save the mouse coordinates
            mouseX = e.X;
            mouseY = e.Y;
        }

        /// <summary>
        /// Used when the mouse button is release (clicked)
        /// </summary>
        private void OnMouseClick(object sender, MouseEventArgs e)
        {
            // If they clicked the mouse button, fire
            RaiseFireEvent();
        }
        #endregion

        #region Network methods/properties
        /// <summary>
        /// Used to determine if the remote player is connected 
        /// </summary>
        public bool IsPlayerConnected
        {
            get { return isPlayerActive; }
            set { isPlayerActive = value; }
        }

        /// <summary>
        /// Is the player the local player
        /// </summary>
        public bool IsLocal
        {
            get { return isPlayerLocal; }
        }


        /// <summary>
        /// Name of the player
        /// </summary>
        public string Name
        {
            get { return playerName; }
            set { playerName = value; }
        }
        #endregion

        #region Device event handlers
        /// <summary>
        /// Fired right before the device is reset and about to be lost.
        /// </summary>
        private void OnDeviceLost(object sender, EventArgs e)
        {
            if ( (playerFont != null) && (!playerFont.Disposed) )
            {
                playerFont.OnLostDevice();
                playerFontSprite.OnLostDevice();
            }
        }
        /// <summary>
        /// Fired when a device is reset
        /// </summary>
        private void OnDeviceReset(object sender, EventArgs e)
        {
            if (playerFont != null)
            {
                playerFont.OnResetDevice();
                playerFontSprite.OnResetDevice();
            }
        }
        #endregion
 
        #region IDisposable Members
        /// <summary>
        /// Cleanup the player here
        /// </summary>
        public void Dispose()
        {
            // Cleanup the player stuff
            if (playerFontSprite != null)
            {
                playerFontSprite.Dispose();
                playerFont.Dispose();
            }

            // Cleanup the ship
            if (myShip != null)
            {
                myShip.Dispose();
            }

        }

        #endregion

		
    }

	/// <summary>
	/// A collection of players
	/// </summary>
	public class Players : IEnumerable
	{
		#region Instance Data
		private ArrayList playerList = new ArrayList();
		#endregion

		#region Collection Methods
		/// <summary>
		/// Indexer for players
		/// </summary>
		public Player this[int index]
		{
			get { return playerList[index] as Player; }
			set { playerList[index] = value; }
		}

		/// <summary>
		/// Add a new player to the list
		/// </summary>
		public void Add(Player b)
		{
			playerList.Add(b);
		}
		/// <summary>
		/// Remove a player to the list
		/// </summary>
		public void Remove(Player b)
		{
			playerList.Remove(b);
		}

		/// <summary>
		/// Remove all players from the list
		/// </summary>
		public void Clear()
		{
			playerList.Clear();
		}

		/// <summary>
		/// The number of items in the list
		/// </summary>
		public int Count
		{
			get { return playerList.Count; }
		}
		#endregion

		#region IEnumerable Members
		/// <summary>
		/// Returns the enumerator for the player list
		/// </summary>
		public IEnumerator GetEnumerator()
		{
			return playerList.GetEnumerator();
		}
		#endregion
	}
}