using System;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Microsoft.Samples.DirectX.UtilityToolkit;

namespace Demo
{
    /// <summary>
    /// Information about a particle
    /// </summary>
    public struct Particle 
    {
        public Vector3 positionVector;       // Current position
        public Vector3 velocityVector;       // Current velocity
        public Vector3 initialPosition;      // Initial position
        public Vector3 initialVelocity;      // Initial velocity
        public float creationTime;     // Time of creation
        public ColorValue diffuseColor; // Initial diffuse color
        public ColorValue fadeColor;    // Faded diffuse color
        public float fadeProgression;      // Fade progression
    };

    /// <summary>
    /// Our special effects class will use a particle system
    /// </summary>
    public class SpecialEffects : IDisposable
    {
        #region Constant Data
        private static readonly ColorValue WhiteSmoke = new ColorValue(
            (float)System.Drawing.Color.WhiteSmoke.R / 255.0f,
            (float)System.Drawing.Color.WhiteSmoke.G / 255.0f,
            (float)System.Drawing.Color.WhiteSmoke.B / 255.0f,
            1.0f);
        private static readonly ColorValue Black = new ColorValue(
            0.0f, 0.0f, 0.0f, 1.0f);
        private const int MaximumParticles = 1024; // This seems like a nice round number
        private static readonly int VertexBufferSize = CustomVertex.PositionColored.StrideSize * 
            MaximumParticles;
        private static readonly int FlushSize = MaximumParticles / 4;
        private static readonly int FlushLockSize = CustomVertex.PositionColored.StrideSize * FlushSize;
        #endregion

        #region Instance data
        // We will store one particle texture
        private Texture particleTexture = null;

        private float time = 0.0f;

        private int baseParticle = MaximumParticles;
        private int particles = 0;
        private System.Collections.ArrayList particlesList = new System.Collections.ArrayList();
        private System.Collections.ArrayList freeParticles = new System.Collections.ArrayList();
        private System.Random rand = new System.Random();

        // A vertex buffer will be needed to render the particles
        private VertexBuffer vertexBuffer = null;
        #endregion

        /// <summary>
        /// Constructor for creating the special effects object
        /// </summary>
        public SpecialEffects(Device device) 
        {
            System.Diagnostics.Debug.Assert(device != null, 
                "You must pass in a valid device to create this class.");

            // The texture needs to be created once
            particleTexture = TexturePool.CreateTexture(device, "particle.bmp");

            // Hook the device events since the vertex buffer will need to be destroyed
            // and recreated when the device is lost or reset
            device.DeviceLost += new EventHandler(OnDeviceLost);
            device.DeviceReset += new EventHandler(OnDeviceReset);

            // Fire the OnDeviceRest once
            OnDeviceReset(device, EventArgs.Empty);
        }
        
        /// <summary>
        /// Add new 'firing' effect
        /// </summary>
        public void AddFiringEffect(int numberParticlesAdding, Vector3 currentPosition, Vector3 vel) 
        {
            // Add some new firing particles
            int particlesEmit = particles + numberParticlesAdding;

            while((particles < MaximumParticles) && (particles < particlesEmit))
            {
                Particle particle;

                if (freeParticles.Count > 0) 
                {
                    particle = (Particle)freeParticles[0];
                    freeParticles.RemoveAt(0);
                }
                else 
                {
                    particle = new Particle();
                }

                // Emit new particle
                particle.initialPosition = currentPosition;
                particle.initialVelocity = Vector3.Normalize(vel);

                particle.initialVelocity.X += ((float)rand.NextDouble() > 0.5f) ? ((float)rand.NextDouble() * 0.1f) : (((float)rand.NextDouble() * 0.1f) * -1);
                particle.initialVelocity.Y += ((float)rand.NextDouble() > 0.5f) ? ((float)rand.NextDouble() * 0.1f) : (((float)rand.NextDouble() * 0.1f) * -1);
                particle.initialVelocity.Z += ((float)rand.NextDouble() > 0.5f) ? ((float)rand.NextDouble() * 0.1f) : (((float)rand.NextDouble() * 0.1f) * -1);
            	
                particle.initialVelocity.Normalize();
                particle.initialVelocity.Scale(vel.Length() * (float)rand.NextDouble());

                particle.positionVector = particle.initialPosition;
                particle.velocityVector = particle.initialVelocity;

                particle.diffuseColor = WhiteSmoke;
                particle.fadeColor = Black;
                particle.fadeProgression = (float)rand.NextDouble();
                particle.creationTime = time;

                particlesList.Add(particle);
                particles++;
            }
        }

        /// <summary>
        /// Updates the particles in the scene
        /// </summary>
        public void Update(float elapsedTime)
        {
            time += elapsedTime;
            for (int ii = particlesList.Count-1; ii >= 0; ii--)
            {
                Particle p = (Particle)particlesList[ii];
                // Calculate new position
                float fT = time - p.creationTime;
                float fGravity;

                fGravity = -290.8f;
                p.fadeProgression -= elapsedTime * 0.25f;

                p.positionVector    = p.initialVelocity * fT + p.initialPosition;
                p.positionVector.Y += (0.5f * fGravity) * (fT * fT);
                p.velocityVector.Y  = p.initialVelocity.Y + fGravity * fT;

                if (p.fadeProgression < 0.0f)
                    p.fadeProgression = 0.0f;

                // Kill old particles
                if (p.fadeProgression <= 0.0f)
                {
                    // Kill particle
                    freeParticles.Add(p);
                    particlesList.RemoveAt(ii);
                    particles--;
                }
                else
                    particlesList[ii] = p;
            }
        }



        /// <summary>
        /// Renders the particles currently in use for the device
        /// </summary>
        public void Draw(GameEngine engine) 
        {
            // Set the render states for using point sprites
            engine.RenderingDevice.RenderState.ZBufferWriteEnable = false;
            engine.RenderingDevice.RenderState.AlphaBlendEnable = true;
            engine.RenderingDevice.RenderState.SourceBlend = Blend.One;
            engine.RenderingDevice.RenderState.DestinationBlend = Blend.One;
            engine.RenderingDevice.RenderState.Lighting = false;

            engine.SetWorldTransform(Matrix.Identity);
            engine.RenderingDevice.RenderState.PointSpriteEnable = true;
            engine.RenderingDevice.RenderState.PointScaleEnable = true;
            engine.RenderingDevice.RenderState.PointSize = Bullet.BulletMass / 2.0f;
            engine.RenderingDevice.RenderState.PointScaleA = 0.0f;
            engine.RenderingDevice.RenderState.PointScaleB = 1.0f;
            engine.RenderingDevice.RenderState.PointScaleC = 1.0f;

            // Set up the vertex buffer to be rendered
            engine.RenderingDevice.SetTexture(0, particleTexture);
            engine.RenderingDevice.SetStreamSource(0, vertexBuffer, 0, CustomVertex.PositionColored.StrideSize);
            engine.RenderingDevice.VertexFormat = CustomVertex.PositionColored.Format;

            int numParticlesToRender = 0;
            // Lock the vertex buffer.  We fill the vertex buffer in small
            // chunks, using LockFlags.NoOverWrite.  When we are done filling
            // each chunk, we call DrawPrim, and lock the next chunk.  When
            // we run out of space in the vertex buffer, we start over at
            // the beginning, using LockFlags.Discard.
            baseParticle += FlushSize;
            if (baseParticle >= MaximumParticles)
            {
                // We've gone too far, reset back to 0
                baseParticle = 0;
            }
            GraphicsStream vertexData = vertexBuffer.Lock(baseParticle * 
                CustomVertex.PositionColored.StrideSize, FlushLockSize, 
                (baseParticle != 0) ? LockFlags.NoOverwrite : LockFlags.Discard);

            foreach(Particle p in particlesList) 
            {
                Vector3 vPos = p.positionVector;
                Vector3 velocityVector = p.velocityVector;
                float LengthSq = velocityVector.LengthSq();
                uint steps;
                if (LengthSq < 1.0f)        steps = 2;
                else if (LengthSq <  4.00f) steps = 3;
                else if (LengthSq <  9.00f) steps = 4;
                else if (LengthSq < 12.25f) steps = 5;
                else if (LengthSq < 16.00f) steps = 6;
                else if (LengthSq < 20.25f) steps = 7;
                else                          steps = 8;
                velocityVector *= -0.01f / (float)steps;
                ColorValue diffuse = ColorOperator.Lerp(p.fadeColor, p.diffuseColor, p.fadeProgression);
                // Render each particle a bunch of times to get a blurring effect
                for (int i = 0; i < steps; i++) 
                {
                    // Write the position and color
                    vertexData.Write(vPos);
                    vertexData.Write(diffuse.ToArgb());
                    if (++numParticlesToRender == FlushSize) 
                    {
                        // Done filling this chunk of the vertex buffer.  Lets unlock and
                        // draw this portion so we can begin filling the next chunk.
                        vertexBuffer.Unlock();
                        engine.RenderingDevice.DrawPrimitives(PrimitiveType.PointList, baseParticle, numParticlesToRender);
                        // Lock the next chunk of the vertex buffer.  If we are at the 
                        // end of the vertex buffer, LockFlags.Discard the vertex buffer and start
                        // at the beginning.  Otherwise, specify LockFlags.NoOverWrite, so we can
                        // continue filling the VB while the previous chunk is drawing.
                        baseParticle += FlushSize;
                        if (baseParticle >= MaximumParticles)
                        {
                            baseParticle = 0;
                        }
                        vertexData = vertexBuffer.Lock(baseParticle * 
                            CustomVertex.PositionColored.StrideSize, FlushLockSize, 
                            (baseParticle != 0) ? LockFlags.NoOverwrite : LockFlags.Discard);
                        
                        numParticlesToRender = 0;
                    }
                    vPos += velocityVector;
                }
            }

            // Unlock the vertex buffer
            vertexBuffer.Unlock();
            // Render any remaining particles
            if (numParticlesToRender > 0)
            {
                engine.RenderingDevice.DrawPrimitives(PrimitiveType.PointList, baseParticle, numParticlesToRender);
            }
            // Reset render states
            engine.RenderingDevice.RenderState.PointSpriteEnable = false;
            engine.RenderingDevice.RenderState.PointScaleEnable = false;
            engine.RenderingDevice.RenderState.Lighting = true;
            engine.RenderingDevice.RenderState.ZBufferWriteEnable = true;
            engine.RenderingDevice.RenderState.AlphaBlendEnable = false;
        }

        #region IDisposable members
        /// <summary>
        /// Clean up the vertex buffer used by the special effects
        /// </summary>
        public void Dispose() 
        {
            // The on device lost method does what we need, use it
            OnDeviceLost(null, EventArgs.Empty);
        }
        #endregion

        #region Device Events
        /// <summary>
        /// Fired when the device is about to be lost
        /// </summary>
        private void OnDeviceLost(object sender, EventArgs e)
        {
            // The vertex buffer simply needs to be destroyed here
            if (vertexBuffer != null)
            {
                vertexBuffer.Dispose();
            }
            vertexBuffer = null;
        }
        /// <summary>
        /// Fired after the device has been reset
        /// </summary>
        private unsafe void OnDeviceReset(object sender, EventArgs e)
        {
            Device device = sender as Device;
            
            System.Diagnostics.Debug.Assert(device != null, 
                "For some reason the event handler has no device.");

            // Create a vertex buffer for the particle system.  The size of this buffer
            // does not relate to the number of particles that exist.  Rather, the
            // buffer is used as a communication channel with the device.. we fill in 
            // a bit, and tell the device to draw.  While the device is drawing, we
            // fill in the next bit using NOOVERWRITE.  We continue doing this until 
            // we run out of vertex buffer space, and are forced to discard the buffer
            // and start over at the beginning.

            vertexBuffer = new VertexBuffer(device, VertexBufferSize, 
                Usage.Dynamic | Usage.WriteOnly | Usage.Points, 
                CustomVertex.PositionColored.Format, Pool.Default);
        }
        #endregion
    }
}