using System;
using System.Drawing;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Microsoft.Samples.DirectX.UtilityToolkit;

namespace Demo
{    
    /// <summary>
    /// The main selection screen for the game
    /// </summary>
    public class MainUiScreen : UiScreen, IDisposable
    {
        #region ScreenState
        /// <summary>
        /// Current state of the user screen
        /// </summary>
        private enum ScreenState
        {
            Main,
            Connection,
            SelectShip,
            SelectHost,
            Loading,
        }
        #endregion

        #region Constants 
        // Button sizes
        private const int ButtonWidth = 208;
        private const int ButtonHeight = 58;
        private const int LoadingWidth = 512;
        private const int LoadingHeight = 48;

        // Texture names
        private const string MainScreenTitleTexture = "title.png";
        private const string ButtonTexture = "Buttons.png";
        private const string UITexture = "dialogs.png";
        #endregion

        #region Instance Data
        private Texture buttonTextures = null;
        private Texture messageTexture = null;
        private Texture dialogTexture = null;

        // Buttons
        private UiButton newButton = null;
        private UiButton exitButton = null;

        // Connection type buttons
        private UiButton hostButton = null;
        private UiButton joinButton = null;

        // Dialog for selecting ship
        private UiButton nextship = null;
        private UiButton prevship = null;
        // Yes button (will be used in more than one spot)
        private UiButton yesButton = null;
        private DialogElement dialogBackground = null;
        private DialogElement selectThisship = null;
        private DialogElement enterHostName = null;
        private UiTextBox hostnameBox = null;

        // Default to the main screen
        private ScreenState currentScreen = ScreenState.Main; 

        // Events
        public event EventHandler HostGame;
        public event EventHandler JoinGame;
        public event EventHandler Quit;

        // Connection information
        private bool isUserHost = false;
        private string host = string.Empty;

        

        #endregion

        #region Creation
        /// <summary>
        /// Create the main menu screen
        /// </summary>
        public MainUiScreen(Device device, int width, int height, GameEngine parent) 
            : base(device, width, height)
        {
            // Create the texture for the background
            messageTexture = TexturePool.CreateTexture(device, MainScreenTitleTexture);

            // Mark the background texture as centered
            StoreTexture(messageTexture, width, height, true);

            // Create the textures for the buttons
            buttonTextures = TexturePool.CreateTexture(device, ButtonTexture);

            // Create the texture for the dialogs
            dialogTexture = TexturePool.CreateTexture(device, UITexture);

            CreateDialogs(device, width, height, parent);
        }

        /// <summary>
        /// Create the dialogs the main screen can have
        /// </summary>
        private void CreateDialogs(Device device, int width, int height, GameEngine parent)
        {
            // Create the new game button
            newButton = new UiButton(renderSprite, buttonTextures, buttonTextures,
                new Rectangle(ButtonWidth,ButtonHeight * 2,
                ButtonWidth, ButtonHeight), 
                new Rectangle(0, ButtonHeight * 2,
                ButtonWidth, ButtonHeight), 
                new Point((width - ButtonWidth) / 2,
                ((height - ButtonHeight) / 2) + (ButtonHeight * 2)),
                parent);

            newButton.Click += new EventHandler(OnNewButton);
            newButton.ShortcutKey = System.Windows.Forms.Keys.Enter;

            // Create the quit game button
            exitButton = new UiButton(renderSprite, buttonTextures, buttonTextures, 
                new Rectangle(ButtonWidth,ButtonHeight * 3,
                ButtonWidth, ButtonHeight), 
                new Rectangle(0, ButtonHeight * 3,
                ButtonWidth, ButtonHeight), 
                new Point((width - ButtonWidth) / 2,
                ((height - ButtonHeight) / 2) + (ButtonHeight * 3)),
                parent);

            exitButton.Click += new EventHandler(OnExitButton);
            exitButton.ShortcutKey = System.Windows.Forms.Keys.Escape;

            // Create the host game button
            hostButton = new UiButton(renderSprite, buttonTextures, buttonTextures,
                new Rectangle(ButtonWidth,ButtonHeight * 0,
                ButtonWidth, ButtonHeight), 
                new Rectangle(0, ButtonHeight * 0,
                ButtonWidth, ButtonHeight), 
                new Point( ((width - ButtonWidth) / 2) - (ButtonWidth / 2),
                ((height - ButtonHeight) / 2) + (ButtonHeight * 2)),
                parent);

            hostButton.Click += new EventHandler(OnHostGame);
            hostButton.ShortcutKey = System.Windows.Forms.Keys.H;

            // Create the join button
            joinButton = new UiButton(renderSprite, buttonTextures, buttonTextures, 
                new Rectangle(ButtonWidth,ButtonHeight * 1,
                ButtonWidth, ButtonHeight), 
                new Rectangle(0, ButtonHeight * 1,
                ButtonWidth, ButtonHeight), 
                new Point( ((width - ButtonWidth) / 2) + (ButtonWidth / 2),
                ((height - ButtonHeight) / 2) + (ButtonHeight * 2)),
                parent);

            joinButton.Click += new EventHandler(OnJoinGame);
            joinButton.ShortcutKey = System.Windows.Forms.Keys.J;

            

            // Create the yes button
            yesButton = new UiButton(renderSprite, buttonTextures, buttonTextures,
                new Rectangle(ButtonWidth,ButtonHeight * 4,
                ButtonWidth, ButtonHeight), 
                new Rectangle(0, ButtonHeight * 4,
                ButtonWidth, ButtonHeight), 
                new Point((width - ButtonWidth) / 2,
                (height - ButtonHeight) - (height / 4 - 20)), parent);

            yesButton.Click += new EventHandler(OnYesButton);
            yesButton.ShortcutKey = System.Windows.Forms.Keys.Enter;

			dialogBackground = new DialogElement(dialogTexture, new Rectangle(0,0, 512, 256), 
				new Point((width - 512) / 2, (height - 256) - 50));

			enterHostName = new DialogElement(dialogTexture, new Rectangle(0,384, 512, 64), 
				new Point((width - 512) / 2, (height-64)/2 + 25));

			hostnameBox = new UiTextBox(device, dialogTexture, new Rectangle(0,448, 512, 64), 
				new Point((width - 512) / 2, (height-64)/2 + 75), parent);
            
            // Setup first button states
            SetButtonStates(true, false, false);
        }

        /// <summary>
        /// Setup button states
        /// </summary>
        private void SetButtonStates(bool isFirstScreen, bool isSecondScreen, bool isSelectHost)
        {
            newButton.IsEnabled = isFirstScreen;
            hostButton.IsEnabled = isSecondScreen;
            joinButton.IsEnabled = isSecondScreen;
            hostnameBox.IsEnabled = isSelectHost;
            yesButton.IsEnabled = isSelectHost;

        }
        #endregion

        #region Rendering
        /// <summary>
        /// Render the main ui screen
        /// </summary>
        public void Draw(GameEngine engine)
        {

            if (currentScreen == ScreenState.Loading)
            {
                // Draw loading screen and nothing more
                DrawLoadingScreen(engine);
                return;
            }

            // Start rendering sprites
            base.BeginSprite();

            // Draw the background
            base.Draw();

            if (currentScreen == ScreenState.Main)
            {
                // Now the buttons
                newButton.Draw();
            }
            else if (currentScreen == ScreenState.Connection)
            {
				System.Threading.Thread.Sleep(400);
                hostButton.Draw();
                joinButton.Draw();
            }

            else if (currentScreen == ScreenState.SelectHost)
            {
                dialogBackground.Draw(renderSprite);
                enterHostName.Draw(renderSprite);
                hostnameBox.Draw(renderSprite, unchecked((int)0xff00AB08));
                yesButton.Draw();
            }

            if (currentScreen != ScreenState.Loading)
            {
                // Draw exit unless the loading screen is on
                exitButton.Draw();
            }

            // You're done rendering sprites
            base.EndSprite();


        }

        /// <summary>
        /// Render the loading screen
        /// </summary>
        public void DrawLoadingScreen(GameEngine engine)
        {
            // Store the loading texture
            StoreTexture(buttonTextures, new Rectangle(0, LoadingWidth - LoadingHeight,
                LoadingWidth, LoadingHeight), screenWidth, screenHeight, true);

            // Start rendering sprites
            base.BeginSprite();
            // Draw the background
            base.Draw();
            // You're done rendering sprites
            base.EndSprite();

            // If we've got here, reset world and view matrices to identity
            engine.RenderingDevice.Transform.World = Matrix.Identity;
            engine.RenderingDevice.Transform.View = Matrix.Identity;
        }
        #endregion

        #region Properties
        /// <summary>
        /// The host name to connect to
        /// </summary>
        public string HostName
        {
            get { return host; }
        }
        /// <summary>
        /// Color of the tank
        /// </summary>

        #endregion

        #region Button Click Event Handlers
        /// <summary>
        /// Fired when the new button is clicked
        /// </summary>
        private void OnNewButton(object sender, EventArgs e)
        {
			System.Threading.Thread.Sleep(800);
            // Go to the next screen
            currentScreen = ScreenState.Connection;
            // Update buttons
            SetButtonStates(false, true, false);
			//System.Threading.Thread.Sleep(400);
        }

        /// <summary>
        /// Fired when the host button is clicked
        /// </summary>
        private void OnHostGame(object sender, EventArgs e)
        {
            // Go to the next screen
            currentScreen = ScreenState.Loading;
            StoreTexture(null, 0, 0, true);
			

            // Update buttons
            SetButtonStates(false, false, false);
            isUserHost = true;
			HostGame(this, e);
        }
        /// <summary>
        /// Fired when the yes button is selected
        /// </summary>
        private void OnYesButton(object sender, EventArgs e)
        {
			System.Threading.Thread.Sleep(400);
            if (currentScreen == ScreenState.SelectHost)
            {
                // Store the host name
                host = hostnameBox.Text;

                // Go to the next screen
                currentScreen = ScreenState.Loading;
                // Update the buttons
                SetButtonStates(false, false, false);
                StoreTexture(null, 0, 0, true);
				JoinGame(this, e);

            } 
            else if (currentScreen == ScreenState.SelectShip)
            {
                // Loading the world now
                currentScreen = ScreenState.Loading;
                // Turn off all buttons
                SetButtonStates(false, false, false);
                exitButton.IsEnabled = false; // Including this one

                // UI is done, fire the appropriate event
                if (isUserHost)
                {
                    // User is hosting
                    if (HostGame != null)
                    {
                        HostGame(this, e);
                    }
                }
                else
                {
                    // User is joining
                    if (JoinGame != null)
                    {
                        JoinGame(this, e);
                    }
                }
            }
        }

        /// <summary>
        /// Fired when the join button is clicked
        /// </summary>
        private void OnJoinGame(object sender, EventArgs e)
        {
            // Go to the next screen
            currentScreen = ScreenState.SelectHost;
            // Update button
            SetButtonStates(false, false, true);
            StoreTexture(null, 0, 0, true);
            isUserHost = false;
			System.Threading.Thread.Sleep(400);
        }





        /// <summary>
        /// Fired when the exit button is clicked
        /// </summary>
        private void OnExitButton(object sender, EventArgs e)
        {
            if (Quit != null)
                Quit(this, e);
        }
        #endregion

        #region IDisposable Members
        /// <summary>
        /// Cleanup any resources
        /// </summary>
        public override void Dispose()
        {
    
            base.Dispose();
        }

        #endregion
    }

}