using System;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace Demo
{
    /// <summary>
    /// The type of camera that is being used
    /// </summary>
    public enum CameraType
    {
        ThirdPersonFar,
        ThirdPersonClose,
        HighView
    }
    /// <summary>
    /// Will hold camera data
    /// </summary>
    public class Camera
    {
        #region Instance Data
        // Attributes for view matrix
        private Vector3 eyeVector;
        private Vector3 lookAtVector;
        private Vector3 upVector;

        // Matrix for the view transform
        private Matrix  viewMatrixTransform;

        // Attributes for projection matrix
        private float fieldView;         
        private float aspectRatio;
        private float nearPlane;
        private float farPlane;

        // Matrix for the projection transform
        private Matrix  projectionMatrixTransform;

        // View frustum data
        private Vector3[] frustumPoints = new Vector3[8];
        private Plane[] frustumPlanes = new Plane[6];

        // The type of camera
        private CameraType localType = CameraType.ThirdPersonFar;
        #endregion

        #region Various properties of the class
        /// <summary>
        /// The location of the eye for the camera
        /// </summary>
        public Vector3 EyeLocation
        { 
            get{ return eyeVector; } 
        }
        /// <summary>
        /// Where the camera is currently looking
        /// </summary>
        public Vector3 LookAtLocation
        { 
            get{ return lookAtVector; } 
        }
        /// <summary>
        /// The 'Up' direction of the camera
        /// </summary>
        public Vector3 UpDirection
        { 
            get{ return upVector; } 
        }
        /// <summary>
        /// The view matrix transform of the camera
        /// </summary>
        public Matrix ViewMatrix
        { 
            get{ return viewMatrixTransform; } 
        }
        /// <summary>
        /// The projection matrix transform of the camera
        /// </summary>
        public Matrix ProjectionMatrix
        { 
            get{ return projectionMatrixTransform; } 
        }
        #endregion

        /// <summary>
        /// Set the viewing parameters
        /// </summary>
        /// <param name="vEyePt">Location of the camera</param>
        /// <param name="vLookatPt">Where the camera is looking</param>
        /// <param name="vUpVec">Up vector of camera</param>
        public void SetViewParameters( Vector3 vEyePt, Vector3 vLookatPt, Vector3 vUpVec )
        {
            // Set attributes for the view matrix
            eyeVector    = vEyePt;
            lookAtVector = vLookatPt;
            upVector    = vUpVec;

            viewMatrixTransform = Matrix.LookAtLH( eyeVector, lookAtVector, upVector );
            UpdateViewFrustum();
        }




        /// <summary>
        /// Set the projection view parameters
        /// </summary>
        /// <param name="fFOV">Field of view</param>
        /// <param name="fAspect">Aspect ratio</param>
        /// <param name="fNearPlane">Near clipping plane</param>
        /// <param name="fFarPlane">Far clipping plane</param>
        public void SetProjParameters( float fFOV, float fAspect, float fNearPlane, float fFarPlane )
        {
            // Set attributes for the projection matrix
            fieldView        = fFOV;
            aspectRatio     = fAspect;
            nearPlane  = fNearPlane;
            farPlane   = fFarPlane;

            projectionMatrixTransform = Matrix.PerspectiveFovLH( fFOV, fAspect, fNearPlane, fFarPlane );
        }

        /// <summary>
        /// Will update the view transforms to track this player
        /// </summary>
        /// <param name="player">player to track</param>
        /// <param name="distance">Distance the camera should be away</param>
        public void Track(IMoveableObject trackingObject)
        {
            // Calculate where the camera should be
            Vector3 cameraPosition = trackingObject.Position;
            Vector3 lookAtPosition = trackingObject.Position;

            //Calculate the camera parameters based on the current type
            switch(localType)
            {
                case CameraType.ThirdPersonClose:
                    cameraPosition.Add(Vector3.Scale(trackingObject.Direction, 
                        -trackingObject.Radius));
                    lookAtPosition.Add(Vector3.Scale(trackingObject.Direction, 
                        trackingObject.Radius));
                    cameraPosition.Y = trackingObject.Radius;
                    lookAtPosition.Y = trackingObject.Radius / 8.0f;
                    break;
                case CameraType.ThirdPersonFar:
                    cameraPosition.Add(Vector3.Scale(trackingObject.Direction, 
                        -trackingObject.Radius * 2.0f));
                    lookAtPosition.Add(Vector3.Scale(trackingObject.Direction, 
                        trackingObject.Radius * 2.0f));

                    cameraPosition.Y = trackingObject.Position.Y + trackingObject.Radius * .75f;
					//cameraPosition.Z -= 5f;
                    lookAtPosition.Y = trackingObject.Position.Y + trackingObject.Radius / 8.0f;
					
                    break;
                case CameraType.HighView:
                    cameraPosition.Add(Vector3.Scale(trackingObject.Direction, 
                        -trackingObject.Radius * 2.0f));
                    lookAtPosition.Add(Vector3.Scale(trackingObject.Direction, 
                        trackingObject.Radius * 2.0f));
                    cameraPosition.Y = 1400;
                    break;
                default:
                    throw new InvalidOperationException("An invalid camera type is selected.");
            }

            // Finally update the camera
            SetViewParameters(cameraPosition, lookAtPosition, GameEngine.UpDirection);
        }

        /// <summary>
        /// Update the view frustum planes
        /// </summary>
        public void UpdateViewFrustum()
        {
            // First get the inverse viewproj matrix
            Matrix mat = viewMatrixTransform * projectionMatrixTransform;
            mat.Invert();

            // Get the 8 corners of the view frustum
            frustumPoints[0] = new Vector3(-1.0f, -1.0f,  0.0f); // xyz
            frustumPoints[1] = new Vector3( 1.0f, -1.0f,  0.0f); // Xyz
            frustumPoints[2] = new Vector3(-1.0f,  1.0f,  0.0f); // xYz
            frustumPoints[3] = new Vector3( 1.0f,  1.0f,  0.0f); // XYz
            frustumPoints[4] = new Vector3(-1.0f, -1.0f,  1.0f); // xyZ
            frustumPoints[5] = new Vector3( 1.0f, -1.0f,  1.0f); // XyZ
            frustumPoints[6] = new Vector3(-1.0f,  1.0f,  1.0f); // xYZ
            frustumPoints[7] = new Vector3( 1.0f,  1.0f,  1.0f); // XYZ

            for( int i = 0; i < frustumPoints.Length; i++ )
                frustumPoints[i] = Vector3.TransformCoordinate(frustumPoints[i], mat);
            
            // Now calculate the planes
            frustumPlanes[0] = Plane.FromPoints(frustumPoints[0], frustumPoints[1], frustumPoints[2] ); // Near
            frustumPlanes[1] = Plane.FromPoints(frustumPoints[6], frustumPoints[7], frustumPoints[5] ); // Far
            frustumPlanes[2] = Plane.FromPoints(frustumPoints[2], frustumPoints[6], frustumPoints[4] ); // Left
            frustumPlanes[3] = Plane.FromPoints(frustumPoints[7], frustumPoints[3], frustumPoints[5] ); // Right
            frustumPlanes[4] = Plane.FromPoints(frustumPoints[2], frustumPoints[3], frustumPoints[6] ); // Top
            frustumPlanes[5] = Plane.FromPoints(frustumPoints[1], frustumPoints[0], frustumPoints[4] ); // Bottom
        }

        /// <summary>
        /// Should this object be drawn based on the current cameras view frustum?
        /// </summary>
        /// <param name="obj">Object to test</param>
        /// <returns>true if the object is in the frustum, false otherwise</returns>
        public bool ObjectInFrustum(IMoveableObject obj)
        {
            // Check to see if this point is in the plane
            foreach(Plane p in frustumPlanes)
            {
                if (p.A * obj.Position.X + p.B * obj.Position.Y + p.C * obj.Position.Z + p.D <= (-obj.Radius))
                {
                    // The object is not in the view frustum, do not draw it
                    return false;
                }
            }
            // By default, the object is in the frustum
            return true;
        }
    }
}