using System;
using System.Collections;
using System.Configuration;
using System.Drawing;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Direct3D=Microsoft.DirectX.Direct3D;
using Microsoft.Samples.DirectX.UtilityToolkit;

namespace Demo
{
    /// <summary>
    /// Main Game Engine
    /// </summary>
    public class GameEngine : IFrameworkCallback, IDeviceCreation
    {
        #region Events
        public event System.Windows.Forms.KeyEventHandler KeyDown;
        public event System.Windows.Forms.KeyEventHandler KeyUp;
        public event System.Windows.Forms.KeyPressEventHandler KeyPress;
        public event System.Windows.Forms.MouseEventHandler MouseMove;
        public event System.Windows.Forms.MouseEventHandler MouseClick;
        #endregion

        #region Creation methods
        /// <summary>Create a new instance of the class</summary>
        public GameEngine(Framework f) 
        { 
            // Store framework
            sF = f; 

			// Try to create the network engine as well
			network = new NetworkEngine();

			// Hook the events
			network.DestroyedPlayer += new Microsoft.DirectX.DirectPlay.PlayerDestroyedEventHandler(
				OnRemoveRemotePlayer);
			network.NewPlayer += new Microsoft.DirectX.DirectPlay.PlayerCreatedEventHandler(
				OnCreateRemotePlayer);
        }
        /// <summary>
        /// Static constructor used to initialize materials
        /// </summary>
        static GameEngine()
        {
            WhiteMaterial = new Material();
            WhiteMaterial.DiffuseColor = new ColorValue(1.0f, 1.0f, 1.0f, 1.0f);
            WhiteMaterial.AmbientColor = new ColorValue(1.0f, 1.0f, 1.0f, 1.0f);
        }
        #endregion

        // Variables
        private Framework sF = null; // Framework for samples

        #region Constants
        public const string GameName = "Space Game";
        public static readonly string MediaPath = ConfigurationSettings.AppSettings.Get("MediaPath");
        public static Material WhiteMaterial;
        public static readonly Vector3 UpDirection = new Vector3(0.0f, 1.0f, 0.0f);
		public int NUM_AI_PLAYERS = 1;
		private Random random = new Random();
		


        #endregion

        #region Instance Data
		// User Interface stuff
		private bool isMainMenuShowing = true;
		private MainUiScreen mainScreen = null;
		private int AICounter = 0;
		private int CurrentLevel = 1;
		private bool amDead = false;

		//Network stuff
		private NetworkEngine network = null;
		private float networkTime = 0.0f;
		private float minNetworkTime = 0.01f;

        // Camera Stuff
        private Camera sceneCamera = new Camera();
		

		// Player Stuff
		private Player localPlayer = null;
		private Player remotePlayer = null;
		private Players playerList = new Players();

		// Sound stuff
		private SoundEngine sounds = null;
		private bool sky = true;
		private bool skip = false;
		
		// AI stuff
		private AIShips aiList = new AIShips();


		// Bullet stuff
		private Bullets allBullets = new Bullets();
		private AIBullets AIbullets = new AIBullets();

        // Game level stuff
		private Skybox worldSky;
		private bool isWorldLoaded = false;
		private SpecialEffects particleEffects = null;
        private bool canClose = true;
		private bool amHosting;

		private Direct3D.Font myFont = null;
        #endregion

        #region Debug Information
        #if (DEBUG)
        private Direct3D.Font debugFont = null;
		
        public int numberVerts = 0;
        public int numberFaces = 0;
        #endif
        #endregion


		#region general stuff
        public bool IsDeviceAcceptable(Caps caps, 
			Format adapterFormat, 
			Format backBufferFormat, 
			bool windowed)
        {
            // Skip back buffer formats that don't support alpha blending
            if (!Manager.CheckDeviceFormat(caps.AdapterOrdinal, 
				caps.DeviceType, 
				adapterFormat, 
                Usage.QueryPostPixelShaderBlending, 
				ResourceType.Textures, 
				backBufferFormat))

                return false;

            // Skip any device that doesn't support at least a single light
            if (caps.MaxActiveLights == 0)
                return false;

            return true;
        }


        public void ModifyDeviceSettings(DeviceSettings settings, Caps caps)
        {
            // This application is designed to work on a pure device by not using 
            // any get methods, so create a pure device if supported and using HWVP.
            if ( (caps.DeviceCaps.SupportsPureDevice) && 
                ((settings.BehaviorFlags & CreateFlags.HardwareVertexProcessing) != 0 ) )
                settings.BehaviorFlags |= CreateFlags.PureDevice;
        }

        private void OnCreateDevice(object sender, DeviceEventArgs e)
        {
            SurfaceDescription desc = e.BackBufferDescription;
		#if (DEBUG)
            // In debug mode, use this font to render stats
            debugFont = new Direct3D.Font(e.Device, new System.Drawing.Font("Arial", 12));
		#endif

		  myFont = new Direct3D.Font(e.Device, new System.Drawing.Font("Arial", 12));

			mainScreen = new MainUiScreen(e.Device,desc.Width,desc.Height, this);
			mainScreen.HostGame += new EventHandler(OnHostGame);
			mainScreen.JoinGame += new EventHandler(OnJoinGame);
			mainScreen.Quit     += new EventHandler(OnMainQuit);
			

//			sF.Render3DEnvironment();
//
//			e.Device.Lights[0].Type = LightType.Directional;
//			e.Device.Lights[0].DiffuseColor = new ColorValue(1.0f, 1.0f, 1.0f, 1.0f);
//			e.Device.Lights[0].Direction = new Vector3(0,-2,0);
//			e.Device.Lights[0].Enabled = true;
//
//			LoadWorld();
        }

        private void OnResetDevice(object sender, DeviceEventArgs e)
        {
            SurfaceDescription desc = e.BackBufferDescription;
            // Set the transformation matrices

			//Camera stuff 
            sceneCamera.SetProjParameters((float)Math.PI / 3, (float)desc.Width / (float)desc.Height, 
                1.0f, 200000.0f);

            e.Device.Transform.Projection = sceneCamera.ProjectionMatrix;
            e.Device.Transform.View = Matrix.Identity;

            // To ensure things are rendered properly within the hierarchy, normalize the normals
            e.Device.RenderState.NormalizeNormals = true;

            e.Device.Lights[0].Type = LightType.Directional;
            e.Device.Lights[0].DiffuseColor = new ColorValue(1.0f, 1.0f, 1.0f, 1.0f);
            e.Device.Lights[0].Direction = new Vector3(0,-2,0);
            e.Device.Lights[0].Enabled = true;

		#if (DEBUG)
            // Make sure the debug font resets correctly
            debugFont.OnResetDevice();
		#endif

            // Store the caps to check for texture filtering
            Caps caps = sF.DeviceCaps;

            // Turn on texture filtering (First magnify)
            if (caps.TextureFilterCaps.SupportsMagnifyAnisotropic)
            {
                e.Device.SamplerState[0].MagFilter = TextureFilter.Anisotropic;
            }
            else if (caps.TextureFilterCaps.SupportsMagnifyLinear)
            {
                e.Device.SamplerState[0].MagFilter = TextureFilter.Linear;
            }
            // Then minify
            if (caps.TextureFilterCaps.SupportsMinifyAnisotropic)
            {
                e.Device.SamplerState[0].MagFilter = TextureFilter.Anisotropic;
            }
            else if (caps.TextureFilterCaps.SupportsMinifyLinear)
            {
                e.Device.SamplerState[0].MagFilter = TextureFilter.Linear;
            }
            // Then mip
            if (caps.TextureFilterCaps.SupportsMipMapLinear)
            {
                e.Device.SamplerState[0].MipFilter = TextureFilter.Linear;
            }
        }

        private void OnLostDevice(object sender, EventArgs e)
        {
		#if (DEBUG)
            if ( (debugFont != null) && (!debugFont.Disposed) )
            {
                // Make sure the debug font gets rid of any resources
                debugFont.OnLostDevice();
            }
		#endif
        }

		private void OnKeyEvent(System.Windows.Forms.Keys key, bool keyDown, bool altDown)
        {
            if (keyDown)
            {
                if (KeyDown != null)
                    KeyDown(this, new System.Windows.Forms.KeyEventArgs(key));

                if (key == System.Windows.Forms.Keys.Escape || key == System.Windows.Forms.Keys.X)
                {
					OnMainQuit(this, EventArgs.Empty);

                    // Either show the 'quit' screen, or simply quit
                   // if (isMainMenuShowing)
//                    {
//                        // Allow the application to quit now
//                        canClose = true;
//                        sF.CloseWindow();
//                    }
                }
            #if (DEBUG)
                    // Handle some debug cases too
                else if (key == System.Windows.Forms.Keys.F1)
                {
                    // turn on wireframe
                    sF.Device.RenderState.FillMode = FillMode.WireFrame;
                }
                else if (key == System.Windows.Forms.Keys.F2)
                {
                    // turn on solid mode
                    sF.Device.RenderState.FillMode = FillMode.Solid;
                }
                else if (key == System.Windows.Forms.Keys.F3)
                {
                    // turn on point
                    sF.Device.RenderState.FillMode = FillMode.Point;
                }
				else if (key == System.Windows.Forms.Keys.F4)
				{
					// turn off lights
					sF.Device.RenderState.Lighting = false;
					System.Diagnostics.Debug.WriteLine("lights off");
				}
				else if (key == System.Windows.Forms.Keys.F5)
				{
					// turn on lights
					sF.Device.RenderState.Lighting = true;
					System.Diagnostics.Debug.WriteLine("lights on");
				}
				else if (key == System.Windows.Forms.Keys.F6)
				{
					if(sky)
						sky = false;
					else
						sky = true;
				}
            #endif
            }
            else
            {
                if (KeyUp != null)
                    KeyUp(this, new System.Windows.Forms.KeyEventArgs(key));

                if (KeyPress != null)
                    KeyPress(this, new System.Windows.Forms.KeyPressEventArgs((char)key));
            }
        }

        /// <summary>
        /// Hook the mouse events
        /// </summary>
        private void OnMouseEvent(bool leftDown, bool rightDown, bool middleDown,
            bool side1Down, bool side2Down, int wheel, int x, int y)
        {
            if (MouseMove != null)
                MouseMove(this, new System.Windows.Forms.MouseEventArgs(0, 0, x, y, wheel));

            if (leftDown)
                if (MouseClick != null)
                    MouseClick(this, new System.Windows.Forms.MouseEventArgs(0, 0, x, y, wheel));
        }

        /// <summary>Return the device being used to render the game</summary>
        public Device RenderingDevice 
		{ 
			get 
			{ 
				return sF.Device; 
			} 
		}

		#endregion

        #region Tank/Level Loading Code
        /// <summary>
        /// Will load everything the world needs
        /// </summary>
        private void LoadWorld(bool isHost)
		{
			System.Threading.Thread.Sleep(2000);
            // Create the world skybox
            worldSky = new Skybox(sF.Device);
            // Create the levels

			if(isHost == true)
			{
				localPlayer = new Player(sF.Device, this, sF.Window, Environment.UserName, true, "viper4.x", Color.DarkSlateBlue);
				localPlayer.Position = new Vector3(500,0,500);
				localPlayer.Rotation=(float)Math.PI;
				
				localPlayer.Fired += new EventHandler(OnPlayerFire);
				//playerList.Add(localPlayer);

				remotePlayer = new Player(sF.Device, this, sF.Window, "Remote", false, "viper4.x", Color.PaleVioletRed);
				remotePlayer.Position = new Vector3(-200,0,-500);
				remotePlayer.Rotation = (float)Math.PI;
				playerList.Add(remotePlayer);
			}
			else
			{
				remotePlayer = new Player(sF.Device, this, sF.Window, "Remote", false, "viper4.x", Color.Cyan);
				remotePlayer.Position = new Vector3(500,0,500);
				remotePlayer.Rotation=(float)Math.PI;
				//Matrix.RotationX((float)Math.PI)* Matrix.RotationY((float)Math.PI);
				remotePlayer.Fired += new EventHandler(OnPlayerFire);
				playerList.Add(remotePlayer);

				localPlayer = new Player(sF.Device, this, sF.Window, Environment.UserName, true, "viper4.x", Color.PaleVioletRed);
				localPlayer.Position = new Vector3(-200,0,-500);
				localPlayer.Rotation = (float)Math.PI;
				//playerList.Add(localPlayer);

			}

			switch (CurrentLevel)
			{
				case 1:

					break;
				case 2:
					NUM_AI_PLAYERS *= 2;
					break;
				case 3:
					NUM_AI_PLAYERS *= 4;
					break;
				default:
					break;
			}


			AICounter = NUM_AI_PLAYERS;
			for (int i=0; i < NUM_AI_PLAYERS; i++)
			{
				AI aiShips = new AI(sF.Device,this,sF.Window,"AI "+i.ToString(),false,"raider.x",Color.PaleVioletRed);
				aiShips.Position = new Vector3(random.Next(-20,20)*1000f,0,random.Next(-20,20)*1000f);
				//aiShips.Position = new Vector3(0,-50,0);
				aiShips.Rotation = (float)Math.PI;
				aiShips.Fired += new EventHandler(OnAIFire);
				aiList.Add(aiShips);
			}

			particleEffects = new SpecialEffects(sF.Device);

			// Now the world is loaded
            isWorldLoaded = true;
        }
        #endregion

        #region Rendering Methods
		/// <summary>
		/// This callback function will be called once at the beginning of every frame. This is the
		/// best location for your application to handle updates to the scene, but is not 
		/// intended to contain actual rendering calls, which should instead be placed in the 
		/// OnFrameRender callback.  
		/// </summary>
		public void OnFrameMove(Device device, double appTime, float elapsedTime)
		{
			#if (DEBUG)
			// Reset the rendering variables in debug mode
			numberVerts = 0;
			numberFaces = 0;
			#endif    
        
			if (isWorldLoaded)
			{
				//send in the clones...er...particles!
				if (particleEffects != null)
				{
					particleEffects.Update(elapsedTime);
				}
					//update the ships
					localPlayer.Update(elapsedTime);


				foreach(Player p in playerList)
				{
					if (p.IsPlayerConnected)
					{
						p.Update(elapsedTime);
					}
					
				}
						
				foreach(AI aiShip in aiList)
				{
					
					if (aiShip.CurrentState == AI.AI_STATE.PATROL)
					{
						// Move ship
						// Scan for new targets
						if (!localPlayer.Dead)
						{
							if (aiShip.ScanForTarget(localPlayer) == true)
							{
								// change status to attack
								// set target to player
								aiShip.SetTarget(localPlayer);
								//System.Diagnostics.Debug.WriteLine("target has been set");
								aiShip.CurrentState=AI.AI_STATE.ATTACK;
							}
							else
							{
								//System.Diagnostics.Debug.WriteLine("patrolling in - scan false");
								foreach (Player p in playerList)
								{
									if (aiShip.ScanForTarget(p) == true)
									{
										// change status to attack
										// set target to player
										aiShip.SetTarget(p);
										//System.Diagnostics.Debug.WriteLine("target has been set");
										aiShip.CurrentState=AI.AI_STATE.ATTACK;
									}
								}

							}
						}

					}
					else // in attack mode
					{
						Player myTarget = (Player)aiShip.GetTarget();
						if (myTarget.Dead)
						{
							aiShip.CurrentState = AI.AI_STATE.PATROL;
							break;
						}
						if (aiShip.Intercept(elapsedTime) == true)
						{
							// already within firing range. Go kill them.
							//System.Diagnostics.Debug.WriteLine("intercept = true");
						}
						else
						{
							//System.Diagnostics.Debug.WriteLine("intercept = false");
						}

					
					}
					aiShip.Update(elapsedTime);					
				}			
			}

			// Make sure only one thread has access to this collection
			lock(allBullets)
			{
				// Update all of the bullets
				for (int i = (allBullets.Count - 1); i >= 0; i--)
				{
					skip = false;
					if (allBullets[i].IsAlive)
					{
						if (allBullets[i].HitTest(localPlayer))
						{
							// Uh oh.. This ship has been exploded
							//sounds.Explosion();
						}
						//if (remotePlayer.IsPlayerConnected)
						{
							foreach(Player player in playerList)
							{
								if (player.IsPlayerConnected)
								{
									if (allBullets[i].HitTest(player))
									{
										// Uh oh.. This ship has been hit
										sounds.Explosion();
									
										player.Armor -= allBullets[i].Damage;

										//System.Diagnostics.Debug.WriteLine("player was hit");
										allBullets.Remove(allBullets[i]);
										//System.Diagnostics.Debug.WriteLine("bullet was removed");
										skip = true;
										break;
								
									}
									if (skip)break;	
								}
							}
							if(skip) break;
							foreach(AI aiShips in aiList)
							{							
								if (allBullets[i].HitTest(aiShips))
								{
									// Uh oh.. This tank has been hit
									sounds.Explosion();
									//System.Diagnostics.Debug.WriteLine(aiShips.Name+" was hit");
									aiShips.Armor -= allBullets[i].Damage;	
									if (aiShips.Armor <= 0) 
									{
										aiList.Remove(aiShips);
										AICounter -= 1;
									}
									allBullets.Remove(allBullets[i]);
									skip = true;
									break;
								}
								if (skip)break;
							}
							if (skip)break;							
						}
						allBullets[i].Update(elapsedTime);
					} // end if
					else //bullet is not alive
					{
						allBullets.Remove(allBullets[i]);
					}
				} // end foreach
			} // end lock
			// Make sure only one thread has access to this collection
			lock(AIbullets)
			{
				// Update all of the bullets
				for (int i = (AIbullets.Count - 1); i >= 0; i--)
				{
					skip = false;
					if (AIbullets[i].IsAlive)
					{
						if(!amDead)
						{
							if (AIbullets[i].HitTest(localPlayer))
							{
								// Uh oh.. This tank has been hit
								sounds.Explosion();
									
								localPlayer.Armor -= AIbullets[i].Damage;

								//System.Diagnostics.Debug.WriteLine("player was hit");
								if (localPlayer.Armor <= 0) 
								{
									//Local Player is dead
									amDead = true;
									localPlayer.Dead = true;

									// draw an explosion!!!

								
								}
								AIbullets.Remove(AIbullets[i]);
								//System.Diagnostics.Debug.WriteLine("bullet was removed");
								skip = true;
								break;
							}
							if (skip)break;	
						}
						//if (remotePlayer.IsPlayerConnected)
					{
						foreach(Player player in playerList)
						{
							if (player.IsPlayerConnected)
							{
								if (AIbullets[i].HitTest(player))
								{
									// Uh oh.. This tank has been hit
									sounds.Explosion();
									
									player.Armor -= AIbullets[i].Damage;

									//System.Diagnostics.Debug.WriteLine("player was hit");
									AIbullets.Remove(AIbullets[i]);
									//System.Diagnostics.Debug.WriteLine("bullet was removed");
									skip = true;
									break;
								
								}
								if (skip)break;	
							}
						}
						if(skip) break;												
					}
						AIbullets[i].Update(elapsedTime);
					}
					else //bullet is not alive
					{
						AIbullets.Remove(AIbullets[i]);
					}
				}
			} // world in not loaded
		}



        /// <summary>
        /// Renders the game scene, including the level, the tanks, etc.
        /// </summary>
        private void RenderGameScene(Device device)
        {
            // Render the skybox first
			if(sky) worldSky.Draw(this, device);
            
            
			sceneCamera.Track(localPlayer);
			// Draw the player
			localPlayer.Draw(this, sceneCamera);
			
			// Now render the player names
			localPlayer.DrawPlayerName(this);
			

			// Is there a remote player to draw?
			if (remotePlayer.IsPlayerConnected)
			{
				remotePlayer.Draw(this, sceneCamera);
				remotePlayer.DrawPlayerName(this);
			}

			foreach(AI aiShips in aiList)
			{
				aiShips.Draw(this, sceneCamera);
				aiShips.DrawPlayerName(this);
			}

			// Make sure only one thread has access to this collection
			lock(allBullets)
			{
				// Draw all of the bullets currently onscreen
				Bullet.SetBulletRenderStates(this);
				foreach(Bullet b in allBullets)
				{
					b.Draw(this, sceneCamera);
				}
				Bullet.EndBulletRenderStates(this);
			}

			lock(AIbullets)
			{
				// Draw all of the bullets currently onscreen
				AIBullet.SetBulletRenderStates(this);
				foreach(AIBullet b in AIbullets)
				{
					b.Draw(this, sceneCamera);
				}
				AIBullet.EndBulletRenderStates(this);
			}

            // Track the local player 
           // 
			//Vector3 eye = new Vector3(-500.0f,00.0f,-500.0f);
			//Vector3 lkat = new Vector3(0.0f,0.0f,1.0f);

			//sceneCamera.SetViewParameters(eye,lkat,UpDirection);

			particleEffects.Draw(this);


        } //end RenderGameScene

        /// <summary>
        /// This callback function will be called at the end of every frame to perform all the 
        /// rendering calls for the scene, and it will also be called if the window needs to be 
        /// repainted. After this function has returned, the sample framework will call 
        /// Device.Present to display the contents of the next buffer in the swap chain
        /// </summary>
        public void OnFrameRender(Device device, double appTime, float elapsedTime)
        {
            bool beginSceneCalled = false;
			bool changeLevel = false;
			bool DeathScreen = false;

            // Clear the render target and the zbuffer 
            device.Clear(ClearFlags.ZBuffer | ClearFlags.Target, 0, 1.0f, 0);
            try
            {
                device.BeginScene();
                beginSceneCalled = true;


				if(isMainMenuShowing)
				{
					mainScreen.Draw(this);
				}
				else if(DeathScreen)
				{
					System.Threading.Thread.Sleep(1000);
					sF.CloseWindow();
					

					
				}
				else if(amDead)					 
				{
					myFont = new Direct3D.Font(device, new System.Drawing.Font("Arial", 22));
					System.Threading.Thread.Sleep(300);
					//local Player is dead
					myFont.DrawText(null, string.Format
						("YOU ARE DEAD!!!!"), new System.Drawing.Rectangle(Framework.DefaultSizeWidth,(int)(Framework.DefaultSizeHeight*0.75),0,0),
						DrawTextFormat.NoClip, unchecked((int)0xffffffff));

					myFont.DrawText(null, string.Format
						("Press Escape to Exit"), new System.Drawing.Rectangle(Framework.DefaultSizeWidth,(int)(Framework.DefaultSizeHeight*1.25),0,0),
						DrawTextFormat.NoClip, unchecked((int)0xff0fffff));
					DeathScreen = true;

				}
				else if(CurrentLevel == 4)
				{
					System.Threading.Thread.Sleep(300);
					//mainScreen.Draw(this);
				}
				else if(!isWorldLoaded)
				{
					LoadWorld(amHosting);
				}
				else if (AICounter > 0)
				{
					
					
					
					
		
					// Set the white material, it will be used for the majority of the rendering
					device.Material = WhiteMaterial;

					// Render the game's scene
					RenderGameScene(device);

					myFont.DrawText(null, string.Format
						("\r\nRemaining Enemies: {0}\r\nArmor: {1}", AICounter,localPlayer.Armor), new System.Drawing.Rectangle(0,0,0,0),
						DrawTextFormat.NoClip, unchecked((int)0xffffffff));
#if (DEBUG)
					debugFont.DrawText(null, string.Format
						("\r\nPosition: {0}\r\nDirection: {1}"
						,localPlayer.Position, localPlayer.Direction), new System.Drawing.Rectangle(2,60,0,0),
						DrawTextFormat.NoClip, unchecked((int)0xffffffff));
#endif
	
				}   
				else if (AICounter == 0)
				{
					if (CurrentLevel < 3)
					{
						myFont = new Direct3D.Font(device, new System.Drawing.Font("Arial", 22));
						myFont.DrawText(null, string.Format
							("ALL ENEMIES DEFEATED"), new System.Drawing.Rectangle(Framework.DefaultSizeWidth,(int)(Framework.DefaultSizeHeight*0.75),0,0),
							DrawTextFormat.NoClip, unchecked((int)0xffff00ff));

						myFont.DrawText(null, string.Format
							("Loading next level "), new System.Drawing.Rectangle(Framework.DefaultSizeWidth,(int)(Framework.DefaultSizeHeight*1.25),0,0),
							DrawTextFormat.NoClip, unchecked((int)0xffffff0f));

					
						CurrentLevel += 1;
						changeLevel = true;
						System.Diagnostics.Debug.WriteLine("Rendering the dead stuff screen");
					}
					else
					{
						myFont.DrawText(null, string.Format
							("YOU HAVE WON THE GAME!!!"), new System.Drawing.Rectangle(Framework.DefaultSizeWidth,(int)(Framework.DefaultSizeHeight*0.75),0,0),
							DrawTextFormat.NoClip, unchecked((int)0xfffffff0));

						myFont.DrawText(null, string.Format
							("Press Escape to Exit"), new System.Drawing.Rectangle(Framework.DefaultSizeWidth,(int)(Framework.DefaultSizeHeight*1.25),0,0),
							DrawTextFormat.NoClip, unchecked((int)0xfff0ffff));
					}
					
				}


                #if (DEBUG)
                // Show debug stats (in debug mode)
                debugFont.DrawText(null, sF.FrameStats, 
					new System.Drawing.Rectangle(2,0,0,0),
                    DrawTextFormat.NoClip, unchecked((int)0xffffff00));
                
				debugFont.DrawText(null, sF.DeviceStats, 
					new System.Drawing.Rectangle(2,15,0,0),
                    DrawTextFormat.NoClip, unchecked((int)0xffffff00));

                // Draw the number of vertices and face information
                debugFont.DrawText(null, string.Format
                    ("\r\nNumber Of Vertices Rendered: {0}\r\nNumber of Faces Rendered: {1}"
                    ,numberVerts, numberFaces), new System.Drawing.Rectangle(2,30,0,0),
                    DrawTextFormat.NoClip, unchecked((int)0xffffffff));

				
                #endif
            }
            finally
            {
                if (beginSceneCalled)
                    device.EndScene();
            }
			if (changeLevel)
			{
				
				isWorldLoaded = false;
			}

        }
        #endregion



        private IntPtr OnWndProc(IntPtr hWnd, NativeMethods.WindowMessage msg, 
			IntPtr wParam, IntPtr lParam, ref bool noFurtherProcessing)
        {
            if (msg == NativeMethods.WindowMessage.Close)
            {
                if (!canClose)
                {
                    noFurtherProcessing = true;
                    return Framework.TrueIntPtr;
                }
            }
            return IntPtr.Zero;
        }

        #region Graphics Methods
        /// <summary>
        /// Will set the world transform and combine the view transform for more efficient rendering
        /// </summary>
        /// <param name="worldTransform">World transformation to apply</param>
        public void SetWorldTransform(Matrix worldTransform)
        {
            sF.Device.Transform.World = worldTransform * sceneCamera.ViewMatrix;
        }
        #endregion

        #region Cleanup

        private void OnDestroyDevice(object sender, EventArgs e)
        {
            // Clean up our objects
            try
            {
    #if (DEBUG)
                // Cleanup the font
                debugFont.Dispose();
    #endif
                // Clean up all textures
                TexturePool.ReleaseAllTextures();

                // Clean up the skybox if it's created
                if (worldSky != null)
                {
                    worldSky.Dispose();
                }
              
            }
            catch
            {
                System.Diagnostics.Debugger.Log(0, GameEngine.GameName, "Error during cleanup.");
            }
        }

        #endregion

        #region Run/Main methods

        /// <summary>
        /// Entry point to the program. Initializes everything and goes into a message processing 
        /// loop. Idle time is used to render the scene.
        /// </summary>
        static int Main() 
        {
            using(Framework sF = new Framework())
            {
                GameEngine Engine = new GameEngine(sF);

                sF.Disposing += new EventHandler(Engine.OnDestroyDevice);
                sF.DeviceLost += new EventHandler(Engine.OnLostDevice);
                sF.DeviceCreated += new DeviceEventHandler(Engine.OnCreateDevice);
                sF.DeviceReset += new DeviceEventHandler(Engine.OnResetDevice);

                sF.SetKeyboardCallback(new KeyboardCallback(Engine.OnKeyEvent));
                // Catch mouse move events
                sF.IsNotifiedOnMouseMove = true;
                sF.SetMouseCallback(new MouseCallback(Engine.OnMouseEvent));

                sF.SetCallbackInterface(Engine);
                try
                {
        #if (!DEBUG)
                    // In retail mode, force the app to fullscreen mode
                    sF.IsOverridingFullScreen = true;
        #endif
                    // Show the cursor and clip it when in full screen
                sF.SetCursorSettings(true, true);

				sF.Initialize( false, false, true );
				// Parse the command line, handle the default hotkeys, and show msgboxes
				sF.CreateWindow("Demo Game Engine");
				sF.CreateDevice( 0, true, Framework.DefaultSizeWidth, 
                        Framework.DefaultSizeHeight, Engine);
					//sF.CreateDevice( 0, true, 1024,	768, Engine);
					

				Engine.sounds = new SoundEngine(sF.Window);

                sF.MainLoop();

                }
        #if(DEBUG)
                catch (Exception e)
                {
                    // In debug mode show this error (maybe - depending on settings)
                    sF.DisplayErrorMessage(e);
        #else
            catch
            {
                // In release mode fail silently
        #endif
                    // Ignore any exceptions here, they would have been handled by other areas
                    return (sF.ExitCode == 0) ? 1 : sF.ExitCode; 
					// Return an error code here
                }

                // Perform any application-level cleanup here. 
				//Direct3D device resources are released within the
                // appropriate callback functions and therefore 
				//don't require any cleanup code here.
                return sF.ExitCode;
            }
        }
        #endregion

		#region Firing Methods
		/// <summary>
		/// The player fired it's weapon
		/// </summary>
		private void OnPlayerFire(object sender, EventArgs e)
		{
			// Play the fire sound
			//sounds.FireWeapon();
			Player p = sender as Player;
			// Make sure only one thread has access to this collection
			lock(allBullets)
			{
				p.FireRound(sF.Device, allBullets);
			}

//			if ((p.IsLocal) && (remotePlayer.IsPlayerConnected))
//			{
//				// Also send this over network
//				network.SendFireMessage();
//			}

			// Fire effect
			if (particleEffects != null)
			{
				
				Vector3 start, vel;
				p.GetBulletStartLocation(out start, out vel);
				// Now add a firing effect to that
				particleEffects.AddFiringEffect(50, start, vel);
			}
		}

		private void OnAIFire(object sender, EventArgs e)
		{
			// Play the fire sound
			//sounds.FireWeapon();
			AI p = sender as AI;
			// Make sure only one thread has access to this collection
			lock(AIbullets)
			{
				p.FireRound(sF.Device, AIbullets);
				System.Diagnostics.Debug.WriteLine("AI round is fired");
			}

			//			if ((p.IsLocal) && (remotePlayer.IsPlayerConnected))
			//			{
			//				// Also send this over network
			//				network.SendFireMessage();
			//			}

			// Fire effect
						if (particleEffects != null)
						{
							// Get the barrel position, and bullet velocity
							Vector3 start, vel;
							p.GetBulletStartLocation(out start, out vel);
							// Now add a firing effect to that
							particleEffects.AddFiringEffect(50, start, vel);
						}
		}
		#endregion

		#region Networking
		private void OnRemoveRemotePlayer(object sender, Microsoft.DirectX.DirectPlay.PlayerDestroyedEventArgs e)
		{
			lock(remotePlayer)
			{
				// The remote player was destroyed.
				remotePlayer.IsPlayerConnected = false;
			}
		}
		private void OnCreateRemotePlayer(object sender, Microsoft.DirectX.DirectPlay.PlayerCreatedEventArgs e)
		{
			lock(remotePlayer)
			{
				// Hey, the remote player was created
				remotePlayer.IsPlayerConnected = true;
				// Tell the remote player what color i am
				//network.SendNewColor(localPlayer.Color);
			}
		}
		#endregion

        #region Main UI Screen Events
        /// <summary>
        /// The user is going to play now and be the host, load everything
        /// </summary>
        private void OnHostGame(object sender, EventArgs e)
        {
            // Get the screen and draw it so the loading screen shows up
            MainUiScreen screen = sender as MainUiScreen;
            sF.Render3DEnvironment();
			amHosting = true;

            // Load the world
            LoadWorld(true);

			network.Host(); 

			isMainMenuShowing = false;
        }

        /// <summary>
        /// The user is ready to join a game now
        /// </summary>
        private void OnJoinGame(object sender, EventArgs e)
        {
            // Get the screen and draw it so the loading screen shows up
            MainUiScreen screen = sender as MainUiScreen;
            sF.Render3DEnvironment();

            // Load the world
            LoadWorld(false);

            // Now join
            network.Join(screen.HostName);

            // We're done with the main menu
            isMainMenuShowing = false;
        }

        /// <summary>
        /// The user wants to quit, let them
        /// </summary>
        private void OnMainQuit(object sender, EventArgs e)
        {
            // Allow the application to quit now
            canClose = true;
            sF.CloseWindow();
        }
        #endregion
    }

    /// <summary>
    ///  The interface used to define any objects that are moveable (ie, player, bullets, etc)
    /// </summary>
    public interface IMoveableObject
    {
        float Radius { get; } 
		// The radius of the moveable object
        Vector3 Position { get; } 
		// The current position of the moveable object
        Vector3 Direction { get; } 
		// The direction the moveable object is facing or moving
        void Update(float elapsedTime); 
		// Update the moveable object
        void Draw(GameEngine engine, Camera c);
		// Render the moveable object
    }
}
