using System;
using System.Configuration;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Direct3D = Microsoft.DirectX.Direct3D;

namespace Demo
{
    /// <summary>
    /// Will contain a UI Screen.
    /// </summary>
    public abstract class UiScreen : IDisposable
    {
        #region Constants
        // Constants
        public static readonly Vector3 ObjectCenter = new Vector3();
        public static readonly int SpriteColor = unchecked((int)0xffffffff);

        #endregion

        #region Instance Data

        protected Sprite renderSprite = null;
        protected Texture backgroundTexture = null;
        protected Rectangle backgroundSource;

        // Screen size
        protected int screenWidth = 0;
        protected int screenHeight = 0;

        // Should the object be centered?
        private bool isCentered = false;
        protected Vector3 centerUpper;
        #endregion

        #region Creation
        /// <summary>
        /// Create a new UI Screen
        /// </summary>
        public UiScreen(Device device, int width, int height)
        {
            // Create the sprite object
            renderSprite = new Sprite(device);

            // Store the width/height
            screenWidth = width;
            screenHeight = height;

            // Hook the device events to 'fix' the sprite if needed
            device.DeviceLost += new EventHandler(OnDeviceLost);
            device.DeviceReset += new EventHandler(OnDeviceReset);

            StoreTexture(null, width, height, false);
        }

        /// <summary>
        /// Store the texture for the background
        /// </summary>
        protected void StoreTexture(Texture background, int width, int height,
            bool centerTexture)
        {
            Rectangle computedSource = Rectangle.Empty;
            if (background != null)
            {
                // Get the background texture
                using (Surface s = background.GetSurfaceLevel(0))
                {
                    SurfaceDescription desc = s.Description;
                    computedSource = new Rectangle(0, 0,
                        desc.Width, desc.Height);
                }
            }

            StoreTexture(background, computedSource, width, height, centerTexture);
        }
        /// <summary>
        /// Store the texture for the background
        /// </summary>
        protected void StoreTexture(Texture background, Rectangle source, int width, int height,
            bool centerTexture)
        {
            // Store the background texture
            backgroundTexture = background;
            backgroundSource = source;

            // Store the centered texture
            isCentered = centerTexture;

            if (isCentered)
            {
                centerUpper = new Vector3((float)(width - backgroundSource.Width) / 2.0f,
                    (float)(height - backgroundSource.Height) / 2.0f, 0.0f);
            }
        }
        #endregion

        #region Device Events
        private void OnDeviceLost(object sender, EventArgs e)
        {
            // The device has been lost, make sure the sprite cleans itself up
            renderSprite.OnLostDevice();
        }

        private void OnDeviceReset(object sender, EventArgs e)
        {
            // The device has been reset, make sure the sprite fixes itself up
            renderSprite.OnResetDevice();
        }
        #endregion

        #region Draw
        /// <summary>
        /// Start drawing with this sprite
        /// </summary>
        protected void BeginSprite()
        {
            renderSprite.Begin(SpriteFlags.AlphaBlend);
        }
        /// <summary>
        /// Staop drawing with this sprite
        /// </summary>
        protected void EndSprite()
        {
            renderSprite.End();
        }
        /// <summary>
        /// Render the button in the correct state
        /// </summary>
        public virtual void Draw()
        {
            // Render the background if it exists
            if (backgroundTexture != null)
            {
                if (isCentered)
                {
                    // Render to the screen centered
                    renderSprite.Draw(backgroundTexture, backgroundSource, 
                        ObjectCenter, centerUpper, SpriteColor);
                }
                else
                {
                    // Scale the background to the right size
                    renderSprite.Transform = Matrix.Scaling(
                        (float)screenWidth / (float)backgroundSource.Width,
                        (float)screenHeight / (float)backgroundSource.Height,
                        0);

                    // Render it to the screen
                    renderSprite.Draw(backgroundTexture, backgroundSource, 
                        ObjectCenter, ObjectCenter, SpriteColor);

                    // Reset the transform
                    renderSprite.Transform = Matrix.Identity;
                }
            }
        }
        #endregion

        #region IDisposable Members
        /// <summary>
        /// Cleanup in case dispose isn't called
        /// </summary>
        ~UiScreen()
        {
            Dispose();
        }
        /// <summary>
        /// Cleanup any resources
        /// </summary>
        public virtual void Dispose()
        {
            GC.SuppressFinalize(this);
            // Dispose the sprite
            if (renderSprite != null)
            {
                renderSprite.Dispose();
            }
            renderSprite = null;
        }

        #endregion
    }

    /// <summary>
    /// Will hold a 'button' that will be rendered via DX
    /// </summary>
    public class UiButton
    {
        #region Instance Data
        private Sprite renderSprite = null;
        private Texture buttonTextureOff = null;
        private Texture buttonTextureOn = null;
        private Rectangle onSource;
        private Rectangle offSource;
        private Vector3 location;
        private bool isButtonOn = false;
        private Rectangle buttonRect;
        private System.Windows.Forms.Keys shortcut;
        private GameEngine parentOwner = null;

        private bool enabled = true; // Should we handle user input?
        // The click event
        public event EventHandler Click;
        #endregion

        #region Creation

        /// <summary>
        /// Create a new instance of the Button class using an existing sprite
        /// </summary>
        public UiButton(Sprite sprite, Texture on, Texture off, 
            Rectangle rectOn, Rectangle rectOff, Point buttonLocation,
            GameEngine parent)
        {
            // Store the sprite object
            renderSprite = sprite;

            // Store the textures
            buttonTextureOff = off;
            buttonTextureOn = on;

            // Rectangles
            onSource = rectOn;
            offSource = rectOff;

            // Location
            location = new Vector3(buttonLocation.X, buttonLocation.Y, 0);

            // Create a rectangle based on the location and size
            buttonRect = new Rectangle((int)location.X, (int)location.Y, 
                onSource.Width, onSource.Height);

            parentOwner = parent;

            HookEvents();
        }

        /// <summary>
        /// Hook (or unhook) the events needed to control the buttons
        /// </summary>
        private void HookEvents()
        {
            if (enabled)
            {
                // Hook the parents events
                parentOwner.MouseClick += new MouseEventHandler(OnMouseClick);
                parentOwner.MouseMove += new MouseEventHandler(OnMouseMove);
                parentOwner.KeyDown += new KeyEventHandler(OnKeyDown);
            }
            else
            {
                // Unhook the parents events
                parentOwner.MouseClick -= new MouseEventHandler(OnMouseClick);
                parentOwner.MouseMove -= new MouseEventHandler(OnMouseMove);
                parentOwner.KeyDown -= new KeyEventHandler(OnKeyDown);
            }
        }
        #endregion

        #region Properties
        /// <summary>
        /// Shortcut key used for this button
        /// </summary>
        public System.Windows.Forms.Keys ShortcutKey
        {
            get { return shortcut; }
            set { shortcut = value; }
        }
        /// <summary>
        /// Is this button enabled?
        /// </summary>
        public bool IsEnabled
        {
            get { return enabled; }
            set { enabled = value; HookEvents(); }
        }
        #endregion

        #region User Input Events
        /// <summary>
        /// Update the button if the mouse is over it
        /// </summary>
        private void OnMouseMove(object sender, MouseEventArgs e)
        {
            if (enabled)
            {
                // Determine if the button is on or not
                isButtonOn = buttonRect.Contains(e.X, e.Y);
            }
        }
        /// <summary>
        /// See if the user clicked the button
        /// </summary>
        private void OnMouseClick(object sender, MouseEventArgs e)
        {
            if (enabled)
            {
                // Determine if the button is pressed
                if(buttonRect.Contains(e.X, e.Y))
                {
                    if (Click != null)
                        Click(this, EventArgs.Empty);
                }
            }
        }

        /// <summary>
        /// Handle any shortcut keys
        /// </summary>
        private void OnKeyDown(object sender, KeyEventArgs e)
        {
            if (enabled)
            {
                // See if the shortcut key was pressed
                if (e.KeyCode == shortcut)
                {
                    // It was, raise the event
                    if (Click != null)
                        Click(this, EventArgs.Empty);
                }
            }
        }
        #endregion

        #region Draw
        /// <summary>
        /// Render the button in the correct state
        /// </summary>
        public void Draw()
        {
            if (isButtonOn)
            {
                renderSprite.Draw(buttonTextureOn, onSource, UiScreen.ObjectCenter, 
                    location, UiScreen.SpriteColor);
            }
            else
            {
                renderSprite.Draw(buttonTextureOff, offSource, UiScreen.ObjectCenter, 
                    location, UiScreen.SpriteColor);
            }
        }
        #endregion
    }

    /// <summary>
    /// Will contain Dialog Elements
    /// </summary>
    public class DialogElement
    {
        #region Instance Data
        private Texture textureSource = null;
        private Rectangle textureRect;
        private Point renderLocation;
        #endregion

        #region Creation
        /// <summary>
        /// Create a new dialog element
        /// </summary>
        public DialogElement(Texture background, Rectangle source, Point loc)
        {
            // Store the information
            textureSource = background;
            textureRect = source;
            renderLocation = loc;
        }
        #endregion

        #region Draw
        /// <summary>
        /// Render the button in the correct state
        /// </summary>
        public virtual void Draw(Sprite renderSprite)
        {
            // Render the element if it exists
            if (textureSource != null)
            {
                // Render to the screen centered
                renderSprite.Draw(textureSource, textureRect, 
                    UiScreen.ObjectCenter, new Vector3(renderLocation.X,
                    renderLocation.Y, 1.0f), UiScreen.SpriteColor);
            }
        }
        #endregion
    }

    /// <summary>
    /// Will contain text boxes
    /// </summary>
    public class UiTextBox
    {
        #region Instance Data
        private Texture textureSource = null;
        private Rectangle textureRect;
        private Rectangle textRectangle;
        private Point renderLocation;
        private Direct3D.Font textFont = null;
        private string currentText = string.Empty;
        private bool enabled = false;
        #endregion

        #region Creation
        /// <summary>
        /// Create a new text box
        /// </summary>
        public UiTextBox(Device device, Texture background, Rectangle source, Point loc, GameEngine parent)
        {
            // Store the information
            textureSource = background;
            textureRect = source;
            renderLocation = loc;
            textRectangle = new Rectangle(loc.X + 20, loc.Y + 15,
                source.Width-40, source.Height - 30);
            // Create the font
            textFont = new Direct3D.Font(device, new System.Drawing.Font(
                "Arial", 20.0f));
            // Hook the device events to handle fonts
            device.DeviceLost += new EventHandler(OnDeviceLost);
            device.DeviceReset += new EventHandler(OnDeviceRest);
            // Hook the keyboard
            parent.KeyPress += new KeyPressEventHandler(OnKeyPress);
        }
        #endregion

        #region Draw
        /// <summary>
        /// Render the button in the correct state
        /// </summary>
        public virtual void Draw(Sprite renderSprite, int color)
        {
            // Render the element if it exists
            if (textureSource != null)
            {
                // Render to the screen centered
                renderSprite.Draw(textureSource, textureRect, 
                    UiScreen.ObjectCenter, new Vector3(renderLocation.X,
                    renderLocation.Y, 1.0f), UiScreen.SpriteColor);

                // Render text
                textFont.DrawText(renderSprite, currentText, textRectangle, 
                    DrawTextFormat.None, color);
            }
        }
        #endregion

        #region Properties
        /// <summary>
        /// Is this button enabled?
        /// </summary>
        public bool IsEnabled
        {
            get { return enabled; }
            set { enabled = value; }
        }
        /// <summary>
        /// Get the current text
        /// </summary>
        public string Text
        {
            get { return currentText; }
        }
        #endregion

        #region Parent events
        /// <summary>
        /// Fired when device is lost
        /// </summary>
        private void OnDeviceLost(object sender, EventArgs e)
        {
            System.Diagnostics.Debug.Assert(textFont != null, "Font cannot be null.");
            textFont.OnLostDevice();
        }

        /// <summary>
        /// Fired when device is reset
        /// </summary>
        private void OnDeviceRest(object sender, EventArgs e)
        {
            System.Diagnostics.Debug.Assert(textFont != null, "Font cannot be null.");
            textFont.OnResetDevice();
        }

        /// <summary>
        /// Used to manipulate text
        /// </summary>
        private void OnKeyPress(object sender, KeyPressEventArgs e)
        {
            if (!enabled)
            {
                // Nothing to do
                return;
            }

            if (e.KeyChar == (char)8)
            {
                if (currentText.Length > 0)
                {
                    // Remove a char
                    currentText = currentText.Substring(0, currentText.Length - 1);
                }
            }
            else
            {
                currentText += e.KeyChar;
            }
        }
        #endregion
    }
}