using System;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace Demo
{
    /// <summary>
    /// This class will maintain the skybox used for the game
    /// </summary>
    public class Skybox : IDisposable
    {
        #region Constants
        private const float ScaleSize = 30000.0f;
        private const float Height = ScaleSize * 2.0f;
        private static readonly Matrix ScaleMatrix = Matrix.Scaling(ScaleSize, ScaleSize, ScaleSize) * 
            Matrix.Translation(0,Height,0);
        private const string SkyboxMeshName = "skybox.x";
        #endregion

        #region Instance Data
        private Mesh skyboxMesh = null;
        private Texture[] skyboxTextures = null;
        #endregion

        #region Creation
        /// <summary>
        /// Create a new instance of the sky box class
        /// </summary>
        public Skybox(Device device)
        {
            // First create the mesh
            ExtendedMaterial[] mtrls = null;
            skyboxMesh = Mesh.FromFile(GameEngine.MediaPath + SkyboxMeshName, MeshFlags.Managed, device,
                out mtrls);

            // With the mesh created, create each of the textures used for the skybox
            if ((mtrls != null) && (mtrls.Length > 0))
            {
                skyboxTextures = new Texture[mtrls.Length];
                for(int i = 0; i < mtrls.Length; i++)
                {
                    skyboxTextures[i] = TexturePool.CreateTexture(device, mtrls[i].TextureFilename);
                }
            }
        }
        #endregion

        #region Rendering
        /// <summary>
        /// Render the sky box
        /// </summary>
        public void Draw(GameEngine engine, Device device)
        {
#if (DEBUG)
            engine.numberFaces += skyboxMesh.NumberFaces;
            engine.numberVerts += skyboxMesh.NumberVertices;
#endif
            // when rendering the skybox you will want to turn off the zbuffer.  
            device.RenderState.ZBufferEnable = false;
            device.RenderState.ZBufferWriteEnable = false;

            // Scale the world transform and turn off lighting
            engine.SetWorldTransform(ScaleMatrix);
            device.RenderState.Lighting = false;
            
            // Render each subset of the mesh
            for(int i = 0; i < skyboxTextures.Length; i++)
            {
                device.SetTexture(0, skyboxTextures[i]);
                skyboxMesh.DrawSubset(i);
            }

            // Turn the zbuffer back on
            device.RenderState.ZBufferEnable = true;
            device.RenderState.ZBufferWriteEnable = true;
            // Turn lighting back on
            device.RenderState.Lighting = true;
        }
        #endregion

        #region IDisposable Members
        /// <summary>
        /// Will be used to cleanup objects
        /// </summary>
        public void Dispose()
        {
            if (skyboxMesh != null)
            {
                skyboxMesh.Dispose();
                skyboxMesh = null;
            }
        }

        #endregion
    }
}